# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License

import pytest
import ezdxf
from ezdxf.math import Vec2
from ezdxf.addons.drawing import RenderContext, Frontend
from ezdxf.addons.drawing.properties import BackendProperties, Properties

from ezdxf.addons.drawing.backend import Backend, BkPath2d
from ezdxf.addons.drawing.debug_backend import BasicBackend, PathBackend
from ezdxf.entities import DXFGraphic
from ezdxf.layouts import Modelspace
from ezdxf.render.forms import cube


class MyTestFrontend(Frontend):
    def __init__(self, ctx, backend):
        super().__init__(ctx, backend)
        self.out = backend


@pytest.fixture
def doc():
    d = ezdxf.new()
    d.layers.new("Test1")
    d.styles.add("DEJAVU", font="DejaVuSans.ttf")
    return d


@pytest.fixture
def msp(doc):
    return doc.modelspace()


@pytest.fixture
def ctx(doc):
    return RenderContext(doc)


@pytest.fixture
def basic(doc, ctx):
    return MyTestFrontend(ctx, BasicBackend())


@pytest.fixture
def path_backend(doc, ctx):
    return MyTestFrontend(ctx, PathBackend())


def unique_types(result):
    return {e[0] for e in result}


def get_result(frontend):
    return frontend.out.collector


def test_basic_frontend_init(basic):
    assert isinstance(basic.out, BasicBackend)


def test_backend_default_draw_path():
    backend = BasicBackend()
    path = BkPath2d.from_vertices(Vec2.list([(0, 0), (1, 0), (2, 0)]))
    backend.draw_path(path, BackendProperties())
    result = backend.collector
    assert len(result) == 2
    assert result[0][0] == "line"


def test_draw_layout(msp, basic):
    msp.add_point((0, 0))
    msp.add_point((0, 0))
    basic.draw_layout(msp)
    result = get_result(basic)
    assert len(result) == 3
    assert result[0][0] == "bgcolor"
    assert result[1][0] == "point"
    assert result[2][0] == "point"


def test_draw_entities(msp, basic):
    msp.add_point((0, 0))
    msp.add_point((0, 0))

    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 2
    assert result[0][0] == "point"
    assert result[1][0] == "point"


def test_filter_draw_entities(msp, basic):
    def filter_layer_l1(e: DXFGraphic) -> bool:
        return e.dxf.layer == "L1"

    msp.add_point((0, 0), dxfattribs={"layer": "L1"})
    msp.add_point((0, 0), dxfattribs={"layer": "L2"})

    basic.draw_entities(msp, filter_func=filter_layer_l1)
    result = get_result(basic)
    assert len(result) == 1
    assert result[0][2].layer == "L1"


def test_point_and_layers(msp, basic):
    msp.add_point((0, 0), dxfattribs={"layer": "Test1"})
    # a non-existing layer shouldn't be a problem
    msp.add_point((0, 0), dxfattribs={"layer": "fail"})
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 2
    assert result[0][0] == "point"
    assert result[0][-1].layer == "Test1"
    assert result[1][0] == "point"
    assert result[1][-1].layer == "fail"


def test_line(msp, basic):
    msp.add_line((0, 0), (1, 0))
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 1
    assert result[0][0] == "line"


def test_lwpolyline_basic(msp, basic):
    msp.add_lwpolyline([(0, 0), (1, 0), (2, 0)])
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 2
    assert unique_types(result) == {"line"}


def test_lwpolyline_path(msp, path_backend):
    msp.add_lwpolyline([(0, 0), (1, 0), (2, 0)])
    path_backend.draw_entities(msp)
    result = get_result(path_backend)
    assert len(result) == 1
    assert unique_types(result) == {"path"}


def test_banded_lwpolyline(msp, basic):
    msp.add_lwpolyline([(0, 0), (1, 0), (2, 0)], dxfattribs={"const_width": 0.1})
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 1
    assert unique_types(result) == {"filled_polygon"}


def test_polyline_2d(msp, basic):
    msp.add_polyline2d([(0, 0), (1, 0), (2, 0)])
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 2
    assert result[0][0] == "line"
    assert result[1][0] == "line"


def test_banded_polyline_2d(msp, basic):
    msp.add_polyline2d(
        [(0, 0, 0.1, 0.2), (1, 0, 0.2, 0.1), (2, 0, 0.1, 0.5)], format="xyse"
    )
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 1
    assert result[0][0] == "filled_polygon"


def test_polyline_3d_basic(msp, basic):
    msp.add_polyline3d([(0, 0, 0), (1, 0, 1), (2, 0, 5)])
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 2
    assert unique_types(result) == {"line"}


def test_polyline_3d_path(msp, path_backend):
    msp.add_polyline3d([(0, 0, 0), (1, 0, 1), (2, 0, 5)])
    path_backend.draw_entities(msp)
    result = get_result(path_backend)
    assert len(result) == 1
    assert unique_types(result) == {"path"}


def test_2d_arc_basic(msp, basic):
    msp.add_circle((0, 0), radius=2)
    msp.add_arc(
        (0, 0),
        radius=2,
        start_angle=30,
        end_angle=60,
        dxfattribs={"layer": "Test1"},
    )
    msp.add_ellipse(
        (0, 0),
        major_axis=(1, 0, 0),
        ratio=0.5,
        start_param=1,
        end_param=2,
        dxfattribs={"layer": "Test1"},
    )
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) > 3
    assert unique_types(result) == {"line"}


def test_3d_circle_basic(msp, basic):
    msp.add_circle((0, 0), radius=2, dxfattribs={"extrusion": (0, 1, 1)})
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) > 30
    assert unique_types(result) == {"line"}


def test_3d_circle_path(msp, path_backend):
    msp.add_circle((0, 0), radius=2, dxfattribs={"extrusion": (0, 1, 1)})
    path_backend.draw_entities(msp)
    result = get_result(path_backend)
    assert len(result) == 1
    assert unique_types(result) == {"path"}


def test_3d_arc_basic(msp, basic):
    msp.add_arc(
        (0, 0),
        radius=2,
        start_angle=30,
        end_angle=60,
        dxfattribs={"extrusion": (0, 1, 1)},
    )
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) >= 4
    assert unique_types(result) == {"line"}


def test_3d_arc_path(msp, path_backend):
    msp.add_arc(
        (0, 0),
        radius=2,
        start_angle=30,
        end_angle=60,
        dxfattribs={"extrusion": (0, 1, 1)},
    )
    path_backend.draw_entities(msp)
    result = get_result(path_backend)
    assert len(result) == 1
    assert unique_types(result) == {"path"}


def test_3d_ellipse_basic(msp, basic):
    msp.add_ellipse(
        (0, 0),
        major_axis=(1, 0, 0),
        ratio=0.5,
        start_param=1,
        end_param=2,
        dxfattribs={"extrusion": (0, 1, 1)},
    )
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) >= 4
    assert unique_types(result) == {"line"}


def test_3d_ellipse_path(msp, path_backend):
    msp.add_ellipse(
        (0, 0),
        major_axis=(1, 0, 0),
        ratio=0.5,
        start_param=1,
        end_param=2,
        dxfattribs={"extrusion": (0, 1, 1)},
    )
    path_backend.draw_entities(msp)
    result = get_result(path_backend)
    assert len(result) == 1
    assert unique_types(result) == {"path"}


def test_2d_text(msp, basic):
    # since v1.0.4 the frontend does the text rendering and passes only filled
    # polygons to the backend
    msp.add_text(
        "test\ntest", dxfattribs={"style": "DEJAVU"}
    )  # \n shouldn't be  problem. Will be ignored
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 8
    assert result[0][0] == "filled_polygon"


def test_ignore_3d_text(msp, basic):
    msp.add_text("test", dxfattribs={"extrusion": (0, 1, 1)})
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 0


def test_mtext(msp, basic):
    # since v1.0.4 the frontend does the text rendering and passes only filled
    # polygons to the backend
    msp.add_mtext("line1\nline2", dxfattribs={"style": "DEJAVU"})
    basic.draw_entities(msp)
    result = get_result(basic)
    assert (
        len(result) == 10
    )  # each character is now one multi-path: changed in v1.1.0b4
    assert result[0][0] == "filled_polygon"


def test_hatch(msp, path_backend):
    hatch = msp.add_hatch()
    hatch.paths.add_polyline_path([(0, 0), (1, 0), (1, 1), (0, 1)])
    path_backend.draw_entities(msp)
    result = get_result(path_backend)
    assert len(result) == 1
    assert result[0][0] == "filled_path"


def test_basic_spline(msp, basic):
    msp.add_spline(fit_points=[(0, 0), (3, 2), (4, 5), (6, 4), (12, 0)])
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) > 1
    entities = {e[0] for e in result}
    assert entities == {"line"}


def test_mesh(msp, basic):
    # draw mesh as wire frame
    c = cube()
    c.render_mesh(msp)
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 24
    assert unique_types(result) == {"line"}


def test_polyface(msp, basic):
    # draw mesh as wire frame
    c = cube()
    c.render_polyface(msp)
    basic.draw_entities(msp)
    result = get_result(basic)
    assert len(result) == 24
    entities = {e[0] for e in result}
    assert entities == {"line"}


class FrontendWithOverride(MyTestFrontend):
    def __init__(self, ctx: RenderContext, out: Backend):
        super().__init__(ctx, out)
        self.override_enabled = True

    def override_properties(self, entity: DXFGraphic, properties: Properties) -> None:
        if not self.override_enabled:
            return
        if properties.layer == "T1":
            properties.layer = "Tx"
        properties.color = "#000000"
        if entity.dxf.text == "T2":
            properties.is_visible = False


def make_override_content(msp: Modelspace):
    msp.delete_all_entities()
    msp.add_text("T0", dxfattribs={"layer": "T0", "color": 7, "style": "DEJAVU"})
    msp.add_text("T1", dxfattribs={"layer": "T1", "color": 6, "style": "DEJAVU"})
    msp.add_text("T2", dxfattribs={"layer": "T2", "color": 5, "style": "DEJAVU"})


def use_override_method(msp: Modelspace, ctx: RenderContext) -> list:
    backend = BasicBackend()
    frontend = FrontendWithOverride(ctx, backend)
    make_override_content(msp)
    frontend.draw_entities(msp)
    frontend.override_enabled = False
    frontend.draw_entities(msp)
    return backend.collector


def override_property_function(entity: DXFGraphic, properties: Properties) -> None:
    if properties.layer == "T1":
        properties.layer = "Tx"
    properties.color = "#000000"
    if entity.dxf.text == "T2":
        properties.is_visible = False


def use_override_function(msp: Modelspace, ctx: RenderContext) -> list:
    backend = BasicBackend()
    frontend = Frontend(ctx, backend)
    make_override_content(msp)

    frontend.push_property_override_function(override_property_function)
    frontend.draw_entities(msp)
    frontend.pop_property_override_function()
    frontend.draw_entities(msp)
    return backend.collector


@pytest.mark.parametrize("override", [use_override_method, use_override_function])
def test_property_override_method(msp: Modelspace, ctx: RenderContext, override):
    collector = override(msp, ctx)

    # since v1.0.4 the frontend does the text rendering and passes only filled
    # polygons to the backend
    assert len(collector) == 10

    # can modify color property
    result = collector[0]
    assert result[0] == "filled_polygon"
    assert result[2].color == "#000000"

    # can modify layer property
    result = collector[2]
    assert result[0] == "filled_polygon"
    assert result[2].layer == "Tx"

    # with override disabled

    result = collector[4]
    assert result[0] == "filled_polygon"
    assert result[2].color == "#ffffff"

    result = collector[6]
    assert result[0] == "filled_polygon"
    assert result[2].layer == "T1"

    result = collector[8]
    assert result[0] == "filled_polygon"
    assert result[2].layer == "T2"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2010-2022, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf
from ezdxf import const
from ezdxf.enums import MTextEntityAlignment
from ezdxf.addons import MTextSurrogate


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new("R12")


def test_horiz_top(doc):
    layout = doc.blocks.new("test_horiz_top")
    text = "lineA\nlineB"
    mtext = MTextSurrogate(text, (0.0, 0.0, 0.0), 1.0)
    mtext.render(layout)
    assert len(layout) == 2
    lines = list(layout)
    assert lines[0].dxftype() == "TEXT"
    assert lines[1].dxftype() == "TEXT"
    assert lines[0].dxf.text == "lineA"
    assert lines[1].dxf.text == "lineB"
    assert lines[0].dxf.align_point == (0, 0, 0)
    assert lines[1].dxf.align_point == (0, -1, 0)
    assert lines[0].dxf.valign == const.TOP
    assert lines[0].dxf.halign == const.LEFT


def test_horiz_bottom(doc):
    layout = doc.blocks.new("test_horiz_bottom")
    text = "lineA\nlineB"
    mtext = MTextSurrogate(
        text, (0.0, 0.0, 0.0), 1.0, align=MTextEntityAlignment.BOTTOM_LEFT
    )
    mtext.render(layout)
    assert len(layout) == 2
    lines = list(layout)
    assert lines[0].dxftype() == "TEXT"
    assert lines[1].dxftype() == "TEXT"
    assert lines[0].dxf.text == "lineA"
    assert lines[1].dxf.text == "lineB"
    assert lines[0].dxf.align_point == (0, 1, 0)
    assert lines[1].dxf.align_point == (0, 0, 0)
    assert lines[0].dxf.valign == const.BOTTOM
    assert lines[0].dxf.halign == const.LEFT


def test_horiz_middle(doc):
    layout = doc.blocks.new("test_horiz_middle")
    text = "lineA\nlineB"
    mtext = MTextSurrogate(
        text, (0.0, 0.0, 0.0), 1.0, align=MTextEntityAlignment.MIDDLE_LEFT
    )
    mtext.render(layout)
    assert len(layout) == 2
    lines = list(layout)
    assert lines[0].dxftype() == "TEXT"
    assert lines[1].dxftype() == "TEXT"
    assert lines[0].dxf.text == "lineA"
    assert lines[1].dxf.text == "lineB"
    assert lines[0].dxf.align_point == (0, 0.5, 0)
    assert lines[1].dxf.align_point == (0, -0.5, 0)
    assert lines[0].dxf.valign == const.MIDDLE
    assert lines[0].dxf.halign == const.LEFT


def test_45deg_top(doc):
    layout = doc.blocks.new("test_45deg_top")
    text = "lineA\nlineB\nlineC"
    mtext = MTextSurrogate(
        text,
        (0.0, 0.0, 0.0),
        1.0,
        align=MTextEntityAlignment.TOP_LEFT,
        rotation=45,
    )
    mtext.render(layout)
    assert len(layout) == 3
    lines = list(layout)
    assert lines[0].dxftype() == "TEXT"
    assert lines[1].dxftype() == "TEXT"
    assert lines[2].dxftype() == "TEXT"
    assert lines[0].dxf.text == "lineA"
    assert lines[1].dxf.text == "lineB"
    assert lines[2].dxf.text == "lineC"
    assert lines[0].dxf.align_point == (0, 0, 0)
    assert lines[1].dxf.align_point == (0.707107, -0.707107, 0)
    assert lines[2].dxf.align_point == (1.414214, -1.414214, 0)
    assert lines[0].dxf.rotation == 45
    assert lines[0].dxf.valign == const.TOP
    assert lines[0].dxf.halign == const.LEFT


def test_45deg_bottom(doc):
    layout = doc.blocks.new("test_45deg_bottom")
    text = "lineA\nlineB\nlineC"
    mtext = MTextSurrogate(
        text,
        (0.0, 0.0, 0.0),
        1.0,
        align=MTextEntityAlignment.BOTTOM_LEFT,
        rotation=45,
    )
    mtext.render(layout)
    assert len(layout) == 3
    lines = list(layout)
    assert lines[0].dxftype() == "TEXT"
    assert lines[1].dxftype() == "TEXT"
    assert lines[2].dxftype() == "TEXT"
    assert lines[0].dxf.text == "lineA"
    assert lines[1].dxf.text == "lineB"
    assert lines[2].dxf.text == "lineC"
    assert lines[0].dxf.align_point == (-1.414214, 1.414214, 0)
    assert lines[1].dxf.align_point == (-0.707107, 0.707107, 0)
    assert lines[2].dxf.align_point == (0, 0, 0)
    assert lines[0].dxf.rotation == 45
    assert lines[0].dxf.valign == const.BOTTOM
    assert lines[0].dxf.halign == const.LEFT


def test_one_liner(doc):
    layout = doc.blocks.new("test_one_liner")
    text = "OneLine"
    mtext = MTextSurrogate(text, (0.0, 0.0, 0.0))
    mtext.render(layout)
    assert len(layout) == 1
    lines = list(layout)
    assert lines[0].dxftype() == "TEXT"
    assert lines[0].dxf.align_point == (0, 0, 0)
    assert lines[0].dxf.valign == const.TOP
    assert lines[0].dxf.halign == const.LEFT


def test_get_attribute_by_subscript():
    mtext = MTextSurrogate("Test\nTest", (0, 0))
    layer = mtext["layer"]
    assert layer == mtext.layer


def test_set_attribute_by_subscript():
    mtext = MTextSurrogate("Test\nTest", (0, 0))
    mtext["layer"] = "modified"
    assert mtext.layer == "modified"



# Copyright (c) 2020-2024, Manfred Moitzi
# License: MIT License
from typing import Iterable
import pytest
import math
import numpy as np

from ezdxf.math.linalg import (
    Matrix,
    detect_banded_matrix,
    compact_banded_matrix,
    BandedMatrixLU,
    banded_matrix,
)
from ezdxf.math.legacy import gauss_vector_solver

BANDED_MATRIX = Matrix(
    matrix=[
        [3, 1, 0, 0, 0, 0, 0],
        [4, 1, 5, 0, 0, 0, 0],
        [9, 2, 6, 5, 0, 0, 0],
        [0, 3, 5, 8, 9, 0, 0],
        [0, 0, 7, 9, 3, 2, 0],
        [0, 0, 0, 3, 8, 4, 6],
        [0, 0, 0, 0, 2, 4, 4],
    ]
)
TRICKY = Matrix(
    matrix=[
        [3, 1, 0, 0, 0, 0, 1],
        [4, 1, 5, 0, 0, 0, 0],
        [9, 2, 6, 5, 0, 0, 0],
        [0, 3, 5, 8, 9, 0, 0],
        [0, 0, 7, 9, 3, 2, 0],
        [0, 0, 0, 3, 8, 4, 6],
        [0, 1, 0, 0, 2, 4, 4],
    ]
)


def are_close_vectors(
    v1: Iterable[float], v2: Iterable[float], abs_tol: float = 1e-12
):
    for i, j in zip(v1, v2):
        assert math.isclose(i, j, abs_tol=abs_tol)


def test_detect_banded_matrix():
    m1, m2 = detect_banded_matrix(BANDED_MATRIX)
    assert (m1, m2) == (2, 1)

    m1, m2 = detect_banded_matrix(TRICKY, check_all=False)
    assert (m1, m2) == (2, 1)

    m1, m2 = detect_banded_matrix(TRICKY, check_all=True)
    assert (m1, m2) == (5, 6)

    assert detect_banded_matrix(Matrix(shape=(10, 10))) == (0, 0)

    identity = Matrix.identity(shape=(10, 10))
    assert detect_banded_matrix(identity) == (0, 0)


def test_compact_banded_matrix():
    m1, m2 = detect_banded_matrix(BANDED_MATRIX)
    m = compact_banded_matrix(BANDED_MATRIX, m1, m2)
    assert m.ncols == 4
    assert m.nrows == 7
    assert m.col(0) == [0, 0, 9, 3, 7, 3, 2]
    assert m.col(1) == [0, 4, 2, 5, 9, 8, 4]
    assert m.col(2) == [3, 1, 6, 8, 3, 4, 4]
    assert m.col(3) == [1, 5, 5, 9, 2, 6, 0]


B1 = [5, 3, 2, 6, 8, 2, 1]
B2 = [9, 1, 7, 6, 4, 5, 0]
B3 = [0, 9, 3, 7, 1, 9, 9]

CHK1 = gauss_vector_solver(BANDED_MATRIX.matrix, B1)
CHK2 = gauss_vector_solver(BANDED_MATRIX.matrix, B2)
CHK3 = gauss_vector_solver(BANDED_MATRIX.matrix, B3)


def test_solve_banded_matrix_vector():
    m, m1, m2 = banded_matrix(BANDED_MATRIX)
    lu = BandedMatrixLU(m, m1, m2)
    are_close_vectors(lu.solve_vector(B1), CHK1)
    are_close_vectors(lu.solve_vector(B2), CHK2)
    are_close_vectors(lu.solve_vector(B3), CHK3)


def test_solve_banded_matrix_matrix():
    m, m1, m2 = banded_matrix(BANDED_MATRIX)
    lu = BandedMatrixLU(m, m1, m2)
    mat_B = np.array(list(zip(B1, B2, B3)))
    r = lu.solve_matrix(mat_B)
    are_close_vectors(r.col(0), CHK1)
    are_close_vectors(r.col(1), CHK2)
    are_close_vectors(r.col(2), CHK3)


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2018 Manfred Moitzi
# License: MIT License
from math import isclose, radians, pi
from ezdxf.math import UCS, Vec3, X_AXIS, Y_AXIS, Z_AXIS, Matrix44


def test_ucs_init():
    ucs = UCS()
    assert ucs.origin == (0, 0, 0)
    assert ucs.ux == (1, 0, 0)
    assert ucs.uy == (0, 1, 0)
    assert ucs.uz == (0, 0, 1)

    assert ucs.from_wcs(Vec3(3, 4, 5)) == (3, 4, 5)
    assert ucs.to_wcs(Vec3(5, 4, 3)) == (5, 4, 3)


def test_ucs_init_ux_uy():
    ucs = UCS(ux=X_AXIS, uy=Y_AXIS)
    assert ucs.uz == Z_AXIS
    ucs = UCS(ux=Y_AXIS, uy=X_AXIS)
    assert ucs.uz == -Z_AXIS


def test_ucs_init_ux_uz():
    ucs = UCS(ux=X_AXIS, uz=Z_AXIS)
    assert ucs.uy == Y_AXIS


def test_ucs_init_uy_uz():
    ucs = UCS(uy=Y_AXIS, uz=Z_AXIS)
    assert ucs.ux == X_AXIS
    ucs = UCS(uz=X_AXIS, uy=Z_AXIS)
    assert ucs.ux == Y_AXIS


def test_translation():
    ucs = UCS(origin=(3, 4, 5))
    assert ucs.origin == (3, 4, 5)
    assert ucs.ux == (1, 0, 0)
    assert ucs.uy == (0, 1, 0)
    assert ucs.uz == (0, 0, 1)
    assert ucs.from_wcs(Vec3(3, 4, 5)) == (0, 0, 0)
    assert ucs.to_wcs(Vec3(1, 1, 1)) == (4, 5, 6)


def test_rotation():
    # normalization is not necessary
    ux = Vec3(1, 2, 0)
    # only cartesian coord systems work
    uy = ux.rotate_deg(90)
    ucs = UCS(ux=ux, uy=uy)
    assert ucs.ux == ux.normalize()
    assert ucs.uy == uy.normalize()
    assert ucs.uz == (0, 0, 1)
    assert ucs.is_cartesian is True


def test_matrix44_rotation():
    # normalization is not necessary
    ux = Vec3(1, 2, 0)
    # only cartesian coord systems work
    uy = ux.rotate_deg(90)
    ucs = UCS(ux=ux, uy=uy)
    m = Matrix44.ucs(ux=ux.normalize(), uy=uy.normalize())
    assert m.ux == ux.normalize()
    assert m.uy == uy.normalize()
    assert m.uz == (0, 0, 1)
    assert m.is_cartesian
    v = m.transform((1, 2, 3))
    assert v == ucs.to_wcs((1, 2, 3))
    assert m.ucs_vertex_from_wcs(v).isclose((1, 2, 3))


def test_transformation():
    axis = Vec3.random()
    angle = 1.5
    ucs = UCS(origin=(3, 4, 5))
    m = Matrix44.axis_rotate(axis, angle)
    expected_origin = m.transform(ucs.origin)
    expected_ux = m.transform(ucs.ux)
    expected_uy = m.transform(ucs.uy)
    expected_uz = m.transform(ucs.uz)
    new = ucs.transform(m)

    assert new.origin.isclose(expected_origin)
    assert new.ux.isclose(expected_ux)
    assert new.uy.isclose(expected_uy)
    assert new.uz.isclose(expected_uz)


def test_none_cartesian():
    ucs = UCS(ux=(1, 2), uy=(0, 2))
    assert ucs.is_cartesian is False


def test_arbitrary_ucs():
    origin = Vec3(3, 3, 3)
    ux = Vec3(1, 2, 0)
    def_point_in_xy_plane = Vec3(3, 10, 4)
    uz = ux.cross(def_point_in_xy_plane - origin)
    ucs = UCS(origin=origin, ux=ux, uz=uz)
    m = Matrix44.ucs(ucs.ux, ucs.uy, ucs.uz, ucs.origin)
    def_point_in_ucs = ucs.from_wcs(def_point_in_xy_plane)

    assert ucs.ux == m.ux
    assert ucs.uy == m.uy
    assert ucs.uz == m.uz
    assert ucs.origin == m.origin

    assert def_point_in_ucs == m.ucs_vertex_from_wcs(def_point_in_xy_plane)
    assert def_point_in_ucs.z == 0
    assert ucs.to_wcs(def_point_in_ucs).isclose(def_point_in_xy_plane)
    assert ucs.is_cartesian is True


def test_ucs_direction_to_ocs_direction():
    ucs = UCS.from_x_axis_and_point_in_xy(
        origin=(1, 2, 3), axis=(2, 3, 4), point=(3, 2, 5)
    )
    assert ucs.is_cartesian is True
    expected = (-3.350073025395333, 2.9626020192591795, 6)
    assert ucs.ucs_direction_to_ocs_direction(Vec3(2, 4, 6)).isclose(expected)


def test_to_ocs():
    ucs = UCS(ux=(0, 0, -1), uz=(1, 0, 0))
    assert ucs.is_cartesian is True
    assert ucs.to_ocs((1, 0, 0)).isclose((0, -1, 0))


def test_points_to_ocs():
    ucs = UCS(ux=(0, 0, -1), uz=(1, 0, 0))
    points = [(1, 2, 3), (4, 5, 6), (9, 8, 7)]
    expected = [ucs.to_ocs(p) for p in points]
    result = list(ucs.points_to_ocs(points))
    assert result == expected


def test_to_ocs_angle_deg():
    ucs = UCS.from_x_axis_and_point_in_xy(
        origin=(1, 2, 3), axis=(2, 3, 4), point=(3, 2, 5)
    )
    expected = 120.077450607124
    assert isclose(ucs.to_ocs_angle_deg(45), expected)


def test_constructor_functions():
    # does not check the math, because this would just duplicate the implementation code
    origin = Vec3(3, 3, 3)
    axis = Vec3(1, 0, -1)
    def_point = Vec3(3, 10, 4)
    ucs = UCS.from_x_axis_and_point_in_xy(origin, axis=axis, point=def_point)
    assert ucs.is_cartesian
    assert isclose(ucs.from_wcs(def_point).z, 0)

    ucs = UCS.from_x_axis_and_point_in_xz(origin, axis=axis, point=def_point)
    assert ucs.is_cartesian
    assert isclose(ucs.from_wcs(def_point).y, 0)

    ucs = UCS.from_y_axis_and_point_in_xy(origin, axis=axis, point=def_point)
    assert ucs.is_cartesian
    assert isclose(ucs.from_wcs(def_point).z, 0)

    ucs = UCS.from_y_axis_and_point_in_yz(origin, axis=axis, point=def_point)
    assert ucs.is_cartesian
    assert isclose(ucs.from_wcs(def_point).x, 0)

    ucs = UCS.from_z_axis_and_point_in_xz(origin, axis=axis, point=def_point)
    assert ucs.is_cartesian
    assert isclose(ucs.from_wcs(def_point).y, 0)

    ucs = UCS.from_z_axis_and_point_in_yz(origin, axis=axis, point=def_point)
    assert ucs.is_cartesian
    assert isclose(ucs.from_wcs(def_point).x, 0)


def test_rotate_x_axis():
    ucs = UCS().rotate((1, 0, 0), radians(90))
    assert ucs.ux.isclose((1, 0, 0))
    assert ucs.uy.isclose((0, 0, 1))
    assert ucs.uz.isclose((0, -1, 0))


def test_rotate_y_axis():
    ucs = UCS().rotate((0, 1, 0), radians(90))
    assert ucs.ux.isclose((0, 0, -1))
    assert ucs.uy.isclose((0, 1, 0))
    assert ucs.uz.isclose((1, 0, 0))


def test_rotate_z_axis():
    ucs = UCS().rotate((0, 0, 1), radians(90))
    assert ucs.ux.isclose((0, 1, 0))
    assert ucs.uy.isclose((-1, 0, 0))
    assert ucs.uz.isclose((0, 0, 1))


def test_rotate_local_x():
    ucs = UCS()
    assert ucs.ux == (1, 0, 0)  # WCS x-axis
    assert ucs.uy == (0, 1, 0)  # WCS y-axis
    assert ucs.uz == (0, 0, 1)  # WCS z-axis
    ucs = ucs.rotate_local_x(pi / 2)
    assert ucs.ux.isclose((1, 0, 0))  # WCS x-axis
    assert ucs.uy.isclose((0, 0, 1))  # WCS z-axis
    assert ucs.uz.isclose((0, -1, 0))  # WCS -y-axis


def test_rotate_local_y():
    ucs = UCS()
    assert ucs.ux == (1, 0, 0)  # WCS x-axis
    assert ucs.uy == (0, 1, 0)  # WCS y-axis
    assert ucs.uz == (0, 0, 1)  # WCS z-axis
    ucs = ucs.rotate_local_y(pi / 2)
    assert ucs.ux.isclose((0, 0, -1))  # WCS -z-axis
    assert ucs.uy.isclose((0, 1, 0))  # WCS y-axis
    assert ucs.uz.isclose((1, 0, 0))  # WCS x-axis


def test_rotate_local_z():
    ucs = UCS()
    assert ucs.ux == (1, 0, 0)  # WCS x-axis
    assert ucs.uy == (0, 1, 0)  # WCS y-axis
    assert ucs.uz == (0, 0, 1)  # WCS z-axis
    ucs = ucs.rotate_local_z(pi / 2)
    assert ucs.ux.isclose((0, 1, 0))  # WCS y-axis
    assert ucs.uy.isclose((-1, 0, 0))  # WCS -x-axis
    assert ucs.uz.isclose((0, 0, 1))  # WCS z-axis


def test_shift_ucs():
    ucs = UCS()
    ucs.shift((1, 2, 3))
    assert ucs.origin == (1, 2, 3)
    ucs.shift((1, 2, 3))
    assert ucs.origin == (2, 4, 6)


def test_moveto():
    ucs = UCS()
    ucs.moveto((1, 2, 3))
    assert ucs.origin == (1, 2, 3)
    ucs.moveto((3, 2, 1))
    assert ucs.origin == (3, 2, 1)



# Copyright (c) 2019-2022, Manfred Moitzi
# License: MIT License
import pytest
import math

import ezdxf
from ezdxf.audit import Auditor
from ezdxf.math import Vec3
from ezdxf.entities.ellipse import Ellipse, MIN_RATIO, MAX_RATIO
from ezdxf.lldxf.tagwriter import TagCollector, basic_tags_from_text

ELLIPSE = """0
ELLIPSE
5
0
330
0
100
AcDbEntity
8
0
100
AcDbEllipse
10
0.0
20
0.0
30
0.0
11
1.0
21
0.0
31
0.0
40
1.0
41
0.0
42
6.283185307179586
"""


@pytest.fixture
def entity():
    return Ellipse.from_text(ELLIPSE)


def test_registered():
    from ezdxf.entities.factory import ENTITY_CLASSES

    assert "ELLIPSE" in ENTITY_CLASSES


def test_default_init():
    entity = Ellipse()
    assert entity.dxftype() == "ELLIPSE"
    assert entity.dxf.handle is None
    assert entity.dxf.owner is None


def test_default_new():
    entity = Ellipse.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "color": 7,
            "ratio": 0.5,
            "center": (1, 2, 3),
            "major_axis": (4, 5, 6),
            "start_param": 10,
            "end_param": 20,
        },
    )
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 7
    assert entity.dxf.center == (1, 2, 3)
    assert entity.dxf.major_axis == (4, 5, 6)
    assert entity.dxf.ratio == 0.5
    assert entity.dxf.start_param == 10
    assert entity.dxf.end_param == 20


def test_extrusion_can_not_be_a_null_vector():
    e = Ellipse.new(dxfattribs={"extrusion": (0, 0, 0)})
    assert e.dxf.extrusion == (0, 0, 1), "expected default extrusion"


def test_major_axis_can_not_be_a_null_vector():
    pytest.raises(ValueError, Ellipse.new, dxfattribs={"major_axis": (0, 0, 0)})


@pytest.mark.parametrize("ratio", [-2, -1, 0, 1, 2])
def test_ratio_is_always_valid(ratio):
    e = Ellipse.new(dxfattribs={"ratio": ratio})
    assert MIN_RATIO <= abs(e.dxf.ratio) <= MAX_RATIO


@pytest.mark.parametrize("ratio", [-1, -0.5, -1e-9])
def test_ratio_can_be_negative(ratio):
    e = Ellipse.new(dxfattribs={"ratio": ratio})
    assert e.dxf.ratio < 0


def test_load_from_text(entity):
    assert entity.dxf.layer == "0"
    assert entity.dxf.color == 256, "default color is 256 (by layer)"
    assert entity.dxf.center == (0, 0, 0)
    assert entity.dxf.major_axis == (1, 0, 0)
    assert entity.dxf.ratio == 1
    assert entity.dxf.start_param == 0
    assert entity.dxf.end_param == math.pi * 2


def test_get_start_and_end_vertex():
    ellipse = Ellipse.new(
        handle="ABBA",
        owner="0",
        dxfattribs={
            "center": (1, 2, 3),
            "major_axis": (4, 3, 0),
            "ratio": 0.7,
            "start_param": math.pi / 2,
            "end_param": math.pi,
            "extrusion": (0, 0, -1),
        },
    )

    start, end = list(
        ellipse.vertices(
            [
                ellipse.dxf.start_param,
                ellipse.dxf.end_param,
            ]
        )
    )
    # test values from BricsCAD
    assert start.isclose(Vec3(3.1, -0.8, 3), abs_tol=1e-6)
    assert end.isclose(Vec3(-3, -1, 3), abs_tol=1e-6)

    # for convenience, but Ellipse.vertices is much more efficient:
    assert ellipse.start_point.isclose(Vec3(3.1, -0.8, 3), abs_tol=1e-6)
    assert ellipse.end_point.isclose(Vec3(-3, -1, 3), abs_tol=1e-6)


def test_write_dxf():
    entity = Ellipse.from_text(ELLIPSE)
    result = TagCollector.dxftags(entity)
    expected = basic_tags_from_text(ELLIPSE)
    assert result == expected


def test_from_arc():
    from ezdxf.entities.arc import Arc

    arc = Arc.new(dxfattribs={"center": (2, 2, 2), "radius": 3})
    ellipse = Ellipse.from_arc(arc)
    assert ellipse.dxf.center == (2, 2, 2)
    assert ellipse.dxf.major_axis == (3, 0, 0)
    assert ellipse.dxf.ratio == 1
    assert ellipse.dxf.start_param == 0
    assert math.isclose(ellipse.dxf.end_param, math.tau)


class TestEllipseParameters:
    @pytest.fixture(scope="class")
    def msp(self):
        doc = ezdxf.new()
        return doc.modelspace()

    def test_adding_ellipse_with_too_big_ratio(self, msp):
        with pytest.raises(ezdxf.DXFValueError):
            msp.add_ellipse(center=(0, 0), major_axis=(1, 0), ratio=2.0)

    def test_adding_ellipse_with_too_small_ratio(self, msp):
        # update: 2022-12-15: min ratio is 1e-10
        ellipse = msp.add_ellipse(center=(0, 0), major_axis=(1, 0), ratio=0.0)
        assert ellipse.dxf.ratio >= 1e-10

    def test_adding_ellipse_with_invalid_major_axis(self, msp):
        with pytest.raises(ezdxf.DXFValueError):
            msp.add_ellipse(center=(0, 0), major_axis=(0, 0), ratio=0.5)

    def test_audit_max_ratio(self, msp):
        ellipse = msp.add_ellipse((0, 0), (1, 0))
        # can only happen for loaded DXF files
        ellipse.dxf.__dict__["ratio"] = 2.0  # hack
        auditor = Auditor(ellipse.doc)
        ellipse.audit(auditor)
        assert len(auditor.fixes) == 1
        assert ellipse.dxf.ratio == 0.5
        assert ellipse.dxf.major_axis.isclose((0.0, 2.0))

    def test_audit_min_ratio(self, msp):
        ellipse = msp.add_ellipse((0, 0), (1, 0))
        # can only happen for loaded DXF files
        ellipse.dxf.__dict__["ratio"] = 1e-11  # hack
        auditor = Auditor(ellipse.doc)
        ellipse.audit(auditor)
        assert len(auditor.fixes) == 1
        # update: 2022-12-15: min ratio is 1e-10
        assert ellipse.dxf.ratio == 1e-10
        assert ellipse.dxf.major_axis.isclose((1.0, 0.0)), "should not changed"

    def test_audit_invalid_major_axis(self, msp):
        ellipse = msp.add_ellipse((0, 0), (1, 0))
        # can only happen for loaded DXF files
        ellipse.dxf.__dict__["major_axis"] = Vec3(0, 0, 0)  # hack
        auditor = Auditor(ellipse.doc)
        ellipse.audit(auditor)
        auditor.empty_trashcan()
        assert len(auditor.fixes) == 1
        assert ellipse.is_alive is False, "invalid ellipse should be deleted"


# tests for swap_axis() are done in test_648_construction_ellipse.py
# tests for params() are done in test_648_construction_ellipse.py

MALFORMED_ELLIPSE = """0
ELLIPSE
5
0
62
7
330
0
6
LT_EZDXF
8
LY_EZDXF
100
AcDbEllipse
10
1.0
20
2.0
30
3.0
100
AcDbEllipse
11
1.0
21
0.0
31
0.0
40
1.0
41
0.0
42
6.283185307179586
"""


def test_malformed_ellipse():
    ellipse = Ellipse.from_text(MALFORMED_ELLIPSE)
    assert ellipse.dxf.layer == "LY_EZDXF"
    assert ellipse.dxf.linetype == "LT_EZDXF"
    assert ellipse.dxf.color == 7
    assert ellipse.dxf.center.isclose((1, 2, 3))
    assert ellipse.dxf.major_axis.isclose((1, 0, 0))
    assert ellipse.dxf.ratio == 1
    assert ellipse.dxf.start_param == 0
    assert ellipse.dxf.end_param == math.pi * 2



# Copyright (c) 2020, Manfred Moitzi
# License: MIT License

import pytest
from ezdxf.render import random_2d_path, random_3d_path


def test_random_2d_path():
    points = list(random_2d_path(steps=100))
    assert len(points) == 100
    assert len(set(points)) > 97


def test_random_3d_path():
    points = list(random_3d_path(steps=100))
    assert len(points) == 100
    assert len(set(points)) > 97


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (C) 2011-2020, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.entities.ltype import Linetype, compile_line_pattern


@pytest.fixture
def linetype():
    ltype = Linetype.new(
        "FFFF",
        dxfattribs={
            "name": "TEST",
            "description": "TESTDESC",
        },
    )
    ltype.setup_pattern([0.2, 0.1, -0.1])
    return ltype


def test_name(linetype):
    assert linetype.dxf.name == "TEST"


def test_description(linetype):
    assert linetype.dxf.description == "TESTDESC"


def test_pattern_items_count(linetype):
    assert isinstance(linetype, Linetype)
    assert len(linetype.pattern_tags) == 7
    assert linetype.pattern_tags.is_complex_type() is False


def test_pattern_tags_details(linetype):
    # pattern tags are accessible but these are implementation details !!!
    assert linetype.pattern_tags.tags[0] == (72, 65)
    assert linetype.pattern_tags.tags[2].value == 0.2


def test_complex_linetype_name():
    complex_ltype = Linetype.new(
        "FFFF",
        dxfattribs={
            "name": "GASLEITUNG",
            "description": "Gasleitung ----GAS----GAS----GAS----GAS----GAS----GAS--",
        },
    )
    complex_ltype.setup_pattern(
        'A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25', 3.0
    )
    assert complex_ltype.dxf.name == "GASLEITUNG"
    assert (
        complex_ltype.dxf.description
        == "Gasleitung ----GAS----GAS----GAS----GAS----GAS----GAS--"
    )
    assert len(complex_ltype.pattern_tags) == 16
    assert (
        complex_ltype.pattern_tags.get_style_handle() == "0"
    ), "Default handle without DXF document"


def test_compile_pattern():
    assert compile_line_pattern(0, [0.0]) == tuple()
    assert compile_line_pattern(2.0, [1.25, -0.25, 0.25, -0.25]) == (
        1.25,
        0.25,
        0.25,
        0.25,
    )
    assert compile_line_pattern(3.5, [2.5, -0.25, 0.5, -0.25]) == (
        2.5,
        0.25,
        0.5,
        0.25,
    )
    assert compile_line_pattern(1.4, [1.0, -0.2, 0.0, -0.2]) == (
        1.0,
        0.2,
        0.0,
        0.2,
    )
    assert compile_line_pattern(0.2, [0.0, -0.2]) == (0.0, 0.2)
    assert compile_line_pattern(2.6, [2.0, -0.2, 0.0, -0.2, 0.0, -0.2]) == (
        2.0,
        0.2,
        0.0,
        0.2,
        0.0,
        0.2,
    )


def test_get_simplified_line_pattern():
    dash_dot = Linetype.new(
        dxfattribs=dict(
            name="DASHDOT",
            description="Dash dot __ . __ . __ . __ . __ . __ . __ . __",
        )
    )
    dash_dot.setup_pattern([1.4, 0.5, 0.5, -0.1, -0.1, 0.0, 0.0, -0.1, -0.1])
    # merges consecutive dashes, gaps and points and returns a
    # line-gap-line-gap... pattern, line == 0 is a point
    pattern = dash_dot.simplified_line_pattern()
    assert pattern == (1.0, 0.2, 0, 0.2)



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License
import math

import pytest
import ezdxf
from ezdxf.document import Drawing
from ezdxf.math import Vec2, arc_angle_span_deg
from ezdxf.render.dim_curved import detect_closer_defpoint, _CurvedDimensionLine


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new(setup=True)


class TestDetectCloserDefpoints:
    @pytest.mark.parametrize(
        "d, offset",  # d=direction
        [
            (Vec2(1, 0), Vec2(0, 0)),  # +x direction
            (Vec2(0, 1), Vec2(1, -1)),  # +y direction
            (Vec2(-1, 0), Vec2(-2, 3)),  # -x direction
            (Vec2(0, -1), Vec2(2, -4)),  # -y direction
            (Vec2(2, -1), Vec2(20, 45)),  # angled
        ],
        ids=["(+x)", "(+y)", "(-x)", "(-y)", "angled"],
    )
    @pytest.mark.parametrize(
        "base",
        [
            0,
            0.5,
            1.0,
            1.5,
            2.0,  # equal -> p1
            -0.5,  # every base left of p1 is closer to p1
            -1.0,
            -100.0,
        ],
    )
    def test_p1_is_closer_to_base(self, base, d, offset):
        # e.g. for base=(-1, 0), d=(1, 0):
        #              base      p1        p2
        # (-x) <---2---(1)---0---(1)---2---(3)---> (+x)
        # By equality p1 if preferred over p2!
        # Shift system by an arbitrary offset!
        p1 = d * 1 + offset
        p2 = d * 3 + offset
        base = d * base + offset
        assert detect_closer_defpoint(d, base, p1, p2) is p1

    @pytest.mark.parametrize(
        "d, offset",  # d=direction
        [
            (Vec2(1, 0), Vec2(0, -1)),  # +x direction
            (Vec2(0, 1), Vec2(2, -2)),  # +y direction
            (Vec2(-1, 0), Vec2(2, 5)),  # -x direction
            (Vec2(0, -1), Vec2(1, 0)),  # -y direction
            (Vec2(2, -1), Vec2(20, 45)),  # angled
        ],
        ids=["(+x)", "(+y)", "(-x)", "(-y)", "angled"],
    )
    @pytest.mark.parametrize(
        "base",
        [
            2.5,
            3.0,
            4.0,  # every base right of p2 is closer to p2
            100.0,
        ],
    )
    def test_p2_is_closer_to_base(self, base, d, offset):
        # e.g. for base=(4.0, 0), d=(1, 0):
        #                      p1        p2    base
        # (-x) <---2---1---0---(1)---2---(3)---(4)---> (+x)
        # By equality p1 if preferred over p2!
        # Shift system by an arbitrary offset!
        p1 = d * 1 + offset
        p2 = d * 3 + offset
        base = d * base + offset
        assert detect_closer_defpoint(d, base, p1, p2) is p2


@pytest.mark.parametrize(
    "s,e",
    [
        [60, 120],
        [300, 240],  # passes 0
        [240, 300],
        [300, 30],  # passes 0
    ],
)
def test_dimension_line_divided_by_measurement_text(doc: Drawing, s, e):
    """Vertical centered measurement text should hide the part of the
    dimension line beneath the text. This creates two arcs instead of one.
    """
    msp = doc.modelspace()
    dim = msp.add_angular_dim_cra(
        center=Vec2(),
        radius=5,
        start_angle=s,
        end_angle=e,
        distance=2,
        override={"dimtad": 0},  # vertical centered text
    )
    dim.render()
    arcs = dim.dimension.get_geometry_block().query("ARC")
    assert len(arcs) == 2
    assert sum(
        arc_angle_span_deg(arc.dxf.start_angle, arc.dxf.end_angle)
        for arc in arcs
    ) < arc_angle_span_deg(
        s, e
    ), "sum of visual arcs should be smaller than the full arc"


def measure_fixed_angle(msp, angle: float):
    x_dist = 15
    radius = 3
    distance = 1
    delta = angle / 2.0
    for dimtad, y_dist in [[0, 0], [1, 20], [4, 40]]:
        for count in range(8):
            center = Vec2(x_dist * count, y_dist)
            main_angle = 45.0 * count
            start_angle = main_angle - delta
            end_angle = main_angle + delta
            yield msp.add_angular_dim_cra(
                center,
                radius,
                start_angle,
                end_angle,
                distance,
                override={"dimtad": dimtad},
            )


def test_text_and_arrows_fit_between_extension_lines(doc: Drawing):
    """There is enough space between extension lines is to place text and
    arrows.
    """
    for dim in measure_fixed_angle(doc.modelspace(), angle=20):
        render_obj = dim.render()
        assert isinstance(render_obj, _CurvedDimensionLine)
        assert render_obj.arrows_outside is False
        assert render_obj.measurement.is_wide_text is False
        assert render_obj.measurement.text_is_outside is False


@pytest.mark.parametrize("angle", [3, 6])
def test_has_outside_text_and_arrows(doc: Drawing, angle):
    """The space between extension lines is too narrow to place text and arrows."""
    for dim in measure_fixed_angle(doc.modelspace(), angle=angle):
        render_obj = dim.render()
        assert isinstance(render_obj, _CurvedDimensionLine)
        assert render_obj.arrows_outside is True
        assert render_obj.measurement.text_is_outside is True
        assert render_obj.measurement.is_wide_text is True


def test_has_outside_text_and_arrows_but_not_a_wide_text(doc: Drawing):
    """The space between extension lines is too narrow to place text and arrows,
    but the text alone has enough space.
    """
    for dim in measure_fixed_angle(doc.modelspace(), angle=9):
        render_obj = dim.render()
        assert isinstance(render_obj, _CurvedDimensionLine)
        assert render_obj.arrows_outside is True
        assert render_obj.measurement.text_is_outside is True
        assert render_obj.measurement.is_wide_text is False


def test_fixed_length_extension_lines(doc: Drawing):
    msp = doc.modelspace()
    dim = msp.add_angular_dim_cra(
        center=(0, 0),
        radius=5,
        distance=2,
        start_angle=0,
        end_angle=90,
        override={
            "dimfxlon": 1,  # use fixed length extension lines
            "dimexe": 0.7,  # length "above" the dimension line
            "dimfxl": 0.5,  # length "below" the dimension line
        },
    ).render()
    # only the extension lines are LINE entities:
    for line in dim.dimension.get_geometry_block().query("LINE"):
        length = line.dxf.start.distance(line.dxf.end)
        assert length == pytest.approx(0.5 + 0.7)


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
import random

import pytest
from ezdxf.addons import genetic_algorithm as ga
from ezdxf.addons import binpacking as bp


class TestFloatDNAZeroOne:
    def test_init_value(self):
        dna = ga.FloatDNA([1.0] * 20)
        assert all(v == 1.0 for v in dna) is True

    @pytest.mark.parametrize("value", [-0.1, 1.1])
    def test_init_value_is_valid(self, value):
        with pytest.raises(ValueError):
            ga.FloatDNA([value] * 20)

    def test_flip_mutate_at(self):
        dna = ga.FloatDNA([0, 0.1, 0.2, 0.3, 0.4])
        for index in range(5):
            dna.flip_mutate_at(index)
        assert list(dna) == pytest.approx([1.0, 0.9, 0.8, 0.7, 0.6])

    def test_iter(self):
        dna = ga.FloatDNA([1.0] * 20)
        assert len(list(dna)) == 20

    def test_reset_data(self):
        dna = ga.FloatDNA([0.0] * 20)
        dna.reset([0.5] * 20)
        assert len(dna) == 20
        assert dna[7] == 0.5

    @pytest.mark.parametrize(
        "values",
        [
            [0, 0, -1],
            [2, 2, 2, 2, 2],
        ],
    )
    def test_reset_data_checks_validity(self, values):
        dna = ga.FloatDNA([])
        with pytest.raises(ValueError):
            dna.reset(values)

    def test_new_random_dna(self):
        dna = ga.FloatDNA.random(20)
        assert len(dna) == 20
        assert len(set(dna)) > 10

    def test_subscription_setter(self):
        dna = ga.FloatDNA([0.0] * 20)
        dna[-3:] = [0.1, 0.2, 0.3]
        assert len(dna) == 20
        assert dna[-3:] == pytest.approx([0.1, 0.2, 0.3])
        assert sum(dna) == pytest.approx(0.6)


class TestBitDNA:
    def test_init_value(self):
        dna = ga.BitDNA([1] * 20)
        assert all(v is True for v in dna) is True

    def test_reset_data(self):
        dna = ga.BitDNA([1] * 20)
        dna.reset([False] * 20)
        assert len(dna) == 20
        assert dna[7] is False

    def test_new_random_dna(self):
        dna = ga.BitDNA.random(20)
        assert len(dna) == 20
        assert len(set(dna)) == 2

    def test_subscription_setter(self):
        dna = ga.BitDNA([1] * 20)
        dna[-3:] = [False, False, False]
        assert len(dna) == 20
        assert dna[-4:] == [True, False, False, False]


class TestUniqueIntDNA:
    def test_init_value(self):
        dna = ga.UniqueIntDNA(10)
        assert dna.is_valid is True
        assert list(dna) == list(range(10))

    def test_init_values(self):
        dna = ga.UniqueIntDNA([0, 1, 2, 3])
        assert dna.is_valid is True
        assert list(dna) == [0, 1, 2, 3]

    def test_init_invalid_values(self):
        with pytest.raises(TypeError):
            ga.UniqueIntDNA([0, 1, 2, 2])

    def test_reset_data(self):
        dna = ga.UniqueIntDNA(10)
        dna.reset(range(9, -1, -1))
        assert len(dna) == 10
        assert dna.is_valid is True
        assert dna[0] == 9
        assert dna[9] == 0

    def test_new_random_dna(self):
        dna = ga.UniqueIntDNA.random(10)
        assert len(dna) == 10
        assert dna.is_valid is True

    def test_subscription_setter(self):
        dna = ga.UniqueIntDNA(10)
        dna[-3:] = [1, 2, 3]
        assert len(dna) == 10
        assert dna[-4:] == [6, 1, 2, 3]
        assert dna.is_valid is False

    def test_recombine_dna_ocx1_preserves_order(self):
        dna1 = ga.UniqueIntDNA(10)  # 0, 1, 2, 3, ...
        dna2 = ga.UniqueIntDNA(10)
        dna2._data.reverse()  # 9, 8, 7, 6, ...
        ga.recombine_dna_ocx1(dna1, dna2, 0, 3)
        assert list(dna1) == [9, 8, 7, 0, 1, 2, 3, 4, 5, 6]
        assert list(dna2) == [0, 1, 2, 9, 8, 7, 6, 5, 4, 3]

    @pytest.mark.parametrize("i1, i2", [(0, 3), (3, 7), (7, 9), (0, 9)])
    def test_random_recombine_dna_ocx1(self, i1, i2):
        dna1 = ga.UniqueIntDNA.random(10)
        dna2 = ga.UniqueIntDNA.random(10)
        copy1 = dna1.copy()
        copy2 = dna2.copy()
        ga.recombine_dna_ocx1(dna1, dna2, i1, i2)
        assert dna1.is_valid is True
        assert dna2.is_valid is True
        assert copy1 != dna1
        assert copy2 != dna2

    @pytest.mark.parametrize("i1, i2", [(0, 0), (8, 8), (9, 9), (10, 11)])
    def test_recombine_dna_ocx1_without_change(self, i1, i2):
        dna1 = ga.UniqueIntDNA.random(10)
        dna2 = ga.UniqueIntDNA.random(10)
        copy1 = dna1.copy()
        copy2 = dna2.copy()
        ga.recombine_dna_ocx1(dna1, dna2, i1, i2)
        assert copy1 == dna1
        assert copy2 == dna2


class TestIntegerDNA:
    def test_init_value(self):
        dna = ga.IntegerDNA([0, 1, 2, 3, 4], 5)
        assert dna.is_valid is True
        assert max(dna) < 5

    def test_init_invalid_data(self):
        with pytest.raises(TypeError):
            ga.IntegerDNA([0, 1, 2, 3, 5], 5)

    def test_flip_mutate_at(self):
        dna = ga.IntegerDNA([0, 1, 2, 3, 4], 5)
        for index in range(5):
            dna.flip_mutate_at(index)
        assert list(dna) == [4, 3, 2, 1, 0]

    def test_reset_data(self):
        dna = ga.IntegerDNA([0, 1, 2, 3, 4, 0, 1, 2, 3, 4], 5)
        dna.reset([4, 3, 2, 1, 0, 1, 2, 3, 2, 1])
        assert len(dna) == 10
        assert dna.is_valid is True
        assert dna[0] == 4
        assert dna[9] == 1

    def test_new_random_dna(self):
        dna = ga.IntegerDNA.random(10, 5)
        assert len(dna) == 10
        assert dna.is_valid is True


class TestHallOfFame:
    @pytest.fixture
    def candidates(self):
        s = []
        for fitness in range(1, 10):
            dna = ga.BitDNA.random(5)
            dna.fitness = 0.1 * fitness
            s.append(dna)
        return s

    def test_build(self, candidates):
        hof = ga.HallOfFame(3)
        for dna in candidates:
            hof.add(dna)
        assert [dna.fitness for dna in hof] == pytest.approx([0.9, 0.8, 0.7])

    def test_get_n_best(self, candidates):
        hof = ga.HallOfFame(3)
        for dna in candidates:
            hof.add(dna)
        result = hof.get(2)
        assert result[0].fitness == pytest.approx(0.9)
        assert result[1].fitness == pytest.approx(0.8)

    def test_get_n_best_negative_values(self, candidates):
        for dna in candidates:
            dna.fitness = -dna.fitness
        hof = ga.HallOfFame(3)
        for dna in candidates:
            hof.add(dna)
        result = hof.get(2)
        assert result[0].fitness == pytest.approx(-0.1)
        assert result[1].fitness == pytest.approx(-0.2)

    def test_purge(self, candidates):
        hof = ga.HallOfFame(3)
        for dna in candidates:
            hof.add(dna)
        hof.purge()
        assert len(hof._unique_entries) == 3


def test_reverse_mutate():
    dna = ga.UniqueIntDNA(10)
    mutate = ga.ReverseMutate(3)
    mutate.mutate(dna, 1.0)
    assert list(dna) != list(ga.UniqueIntDNA(10))
    assert len(set(dna)) == 10


def test_scramble_mutate():
    dna = ga.UniqueIntDNA(10)
    mutate = ga.ScrambleMutate(5)
    mutate.mutate(dna, 1.0)
    assert list(dna) != list(ga.UniqueIntDNA(10))
    assert len(set(dna)) == 10


def test_tournament_selection():
    candidates = [ga.UniqueIntDNA(10) for _ in range(10)]
    for index, dna in enumerate(candidates):
        dna.fitness = index
    selection = ga.TournamentSelection(2)
    selection.reset(candidates)

    result = list(selection.pick(2))
    assert len(result) == 2

    result = list(selection.pick(3))
    assert len(result) == 3


class TestRouletteSelection:
    SELECTOR = ga.RouletteSelection

    @pytest.mark.parametrize("low, high", [
        (1, 10000),
        (-10000, -1),
    ])
    def test_weights(self, low, high):
        dna_max, dna_min = ga.BitDNA.n_random(2, 10)
        dna_max.fitness = high
        dna_min.fitness = low
        selector = self.SELECTOR(negative_values=high < 0)
        selector.reset([dna_max, dna_min])
        random.seed(42)
        count = 0
        for _ in range(4):
            # first runs may fail in test mode, only testing issue?
            # in real world application the relation dna_max:dna_min is always 20:0
            values = list(selector.pick(20))
            # in debug mode this assertion is ALWAYS True (count==4)
            if values.count(dna_max) > values.count(dna_min):
                count += 1
        assert count > 1


class TestRankBasedSelection(TestRouletteSelection):
    SELECTOR = ga.RankBasedSelection


def test_two_point_crossover():
    dna1 = ga.BitDNA([False] * 20)
    dna2 = ga.BitDNA([True] * 20)
    ga.recombine_dna_2pcx(dna1, dna2, 7, 11)
    assert list(dna1[0:7]) == [False] * 7
    assert list(dna1[7:11]) == [True] * 4
    assert list(dna1[11:]) == [False] * 9
    assert list(dna2[0:7]) == [True] * 7
    assert list(dna2[7:11]) == [False] * 4
    assert list(dna2[11:]) == [True] * 9


class TestThresholdFilter:
    @pytest.fixture
    def candidates(self):
        return [ga.BitDNA([]) for _ in range(100)]

    def test_positive_values(self, candidates):
        for fitness, c in enumerate(candidates):
            c.fitness = fitness
        candidates = list(ga.threshold_filter(candidates, 99, 0.1))
        assert len(candidates) == 90

    def test_negative_values(self, candidates):
        for fitness, c in enumerate(candidates):
            c.fitness = -fitness
        candidates = list(ga.threshold_filter(candidates, 0.0, 0.1))
        assert len(candidates) == 90


SMALL_ENVELOPE = ("small-envelope", 11.5, 6.125, 0.25, 10)
LARGE_ENVELOPE = ("large-envelope", 15.0, 12.0, 0.75, 15)
SMALL_BOX = ("small-box", 8.625, 5.375, 1.625, 70.0)
MEDIUM_BOX = ("medium-box", 11.0, 8.5, 5.5, 70.0)
MEDIUM_BOX2 = ("medium-box-2", 13.625, 11.875, 3.375, 70.0)
LARGE_BOX = ("large-box", 12.0, 12.0, 5.5, 70.0)
LARGE_BOX2 = ("large-box-2", 23.6875, 11.75, 3.0, 70.0)

ALL_BINS = [
    SMALL_ENVELOPE,
    LARGE_ENVELOPE,
    SMALL_BOX,
    MEDIUM_BOX,
    MEDIUM_BOX2,
    LARGE_BOX,
    LARGE_BOX2,
]


@pytest.fixture
def packer():
    packer = bp.Packer()
    packer.add_item("50g [powder 1]", 3.9370, 1.9685, 1.9685, 1)
    packer.add_item("50g [powder 2]", 3.9370, 1.9685, 1.9685, 2)
    packer.add_item("50g [powder 3]", 3.9370, 1.9685, 1.9685, 3)
    packer.add_item("250g [powder 4]", 7.8740, 3.9370, 1.9685, 4)
    packer.add_item("250g [powder 5]", 7.8740, 3.9370, 1.9685, 5)
    packer.add_item("250g [powder 6]", 7.8740, 3.9370, 1.9685, 6)
    packer.add_item("250g [powder 7]", 7.8740, 3.9370, 1.9685, 7)
    packer.add_item("250g [powder 8]", 7.8740, 3.9370, 1.9685, 8)
    packer.add_item("250g [powder 9]", 7.8740, 3.9370, 1.9685, 9)
    return packer


def pack(packer, box, pick):
    packer.add_bin(*box)
    packer.pack(pick)
    return packer.bins[0]


class DummyEvaluator(ga.Evaluator):
    def __init__(self, packer: bp.AbstractPacker):
        self.packer = packer

    def evaluate(self, dna: ga.DNA) -> float:
        return 0.5

    def run_packer(self, dna: ga.DNA):
        packer = self.packer.copy()
        packer.pack()
        return packer


class TestGeneticOptimizer:
    def test_init(self, packer):
        driver = ga.GeneticOptimizer(packer, 100)
        assert driver.is_executed is False

    def test_init_invalid_max_runs(self, packer):
        with pytest.raises(ValueError):
            ga.GeneticOptimizer(packer, 0)

    def test_can_only_run_once(self, packer):
        evaluator = DummyEvaluator(packer)
        optimizer = ga.GeneticOptimizer(evaluator, 100)
        optimizer.execute()
        assert optimizer.is_executed is True
        with pytest.raises(TypeError):
            optimizer.execute()

    def test_execution(self, packer):
        packer.add_bin(*MEDIUM_BOX)
        evaluator = DummyEvaluator(packer)
        optimizer = ga.GeneticOptimizer(evaluator, 10)
        optimizer.add_candidates(ga.BitDNA.n_random(20, len(packer.items)))
        optimizer.execute()
        assert optimizer.generation == 10
        assert optimizer.best_fitness > 0.1

        # Get best packer of SubSetEvaluator:
        best_packer = evaluator.run_packer(optimizer.best_dna)
        assert len(best_packer.bins[0].items) > 1


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2019-2022, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Optional
import logging

from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.const import DXF12, SUBCLASS_MARKER, DXF2000, DXF2007, DXF2010
from ezdxf.math import Vec3, NULLVEC
from ezdxf.entities.dxfentity import base_class, SubclassProcessor, DXFEntity
from ezdxf.entities.layer import acdb_symbol_table_record
from .factory import register_entity

if TYPE_CHECKING:
    from ezdxf.entities import DXFNamespace
    from ezdxf.lldxf.tagwriter import AbstractTagWriter

__all__ = ["View"]
logger = logging.getLogger("ezdxf")

acdb_view = DefSubclass(
    "AcDbViewTableRecord",
    {
        "name": DXFAttr(2, validator=validator.is_valid_table_name),
        "flags": DXFAttr(70, default=0),
        "height": DXFAttr(40, default=1),
        "width": DXFAttr(41, default=1),
        "center": DXFAttr(10, xtype=XType.point2d, default=NULLVEC),
        "direction": DXFAttr(
            11,
            xtype=XType.point3d,
            default=Vec3(1, 1, 1),
            validator=validator.is_not_null_vector,
        ),
        "target": DXFAttr(12, xtype=XType.point3d, default=NULLVEC),
        "focal_length": DXFAttr(42, default=50),
        "front_clipping": DXFAttr(43, default=0),
        "back_clipping": DXFAttr(44, default=0),
        "view_twist": DXFAttr(50, default=0),
        "view_mode": DXFAttr(71, default=0),
        # Render mode:
        # 0 = 2D Optimized (classic 2D)
        # 1 = Wireframe
        # 2 = Hidden line
        # 3 = Flat shaded
        # 4 = Gouraud shaded
        # 5 = Flat shaded with wireframe
        # 6 = Gouraud shaded with wireframe
        "render_mode": DXFAttr(
            281,
            default=0,
            dxfversion=DXF2000,
            validator=validator.is_in_integer_range(0, 7),
            fixer=RETURN_DEFAULT,
        ),
        # 1 if there is an UCS associated to this view, 0 otherwise.
        "ucs": DXFAttr(
            72,
            default=0,
            dxfversion=DXF2000,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "ucs_origin": DXFAttr(110, xtype=XType.point3d, dxfversion=DXF2000),
        "ucs_xaxis": DXFAttr(
            111,
            xtype=XType.point3d,
            dxfversion=DXF2000,
            validator=validator.is_not_null_vector,
        ),
        "ucs_yaxis": DXFAttr(
            112,
            xtype=XType.point3d,
            dxfversion=DXF2000,
            validator=validator.is_not_null_vector,
        ),
        # 0 = UCS is not orthographic
        # 1 = Top
        # 2 = Bottom
        # 3 = Front
        # 4 = Back
        # 5 = Left
        # 6 = Right
        "ucs_ortho_type": DXFAttr(
            79,
            dxfversion=DXF2000,
            validator=validator.is_in_integer_range(0, 7),
            fixer=lambda x: 0,
        ),
        "elevation": DXFAttr(146, dxfversion=DXF2000, default=0),
        # handle of AcDbUCSTableRecord if UCS is a named UCS. If not present,
        # then UCS is unnamed:
        "ucs_handle": DXFAttr(345, dxfversion=DXF2000),
        # handle of AcDbUCSTableRecord of base UCS if UCS is orthographic (79 code
        # is non-zero). If not present and 79 code is non-zero, then base UCS is
        # taken to be WORLD
        "base_ucs_handle": DXFAttr(346, dxfversion=DXF2000),
        # 1 if the camera is plottable
        "camera_plottable": DXFAttr(
            73,
            default=0,
            dxfversion=DXF2007,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        "background_handle": DXFAttr(332, optional=True, dxfversion=DXF2007),
        "live_selection_handle": DXFAttr(
            334, optional=True, dxfversion=DXF2007
        ),
        "visual_style_handle": DXFAttr(348, optional=True, dxfversion=DXF2007),
        "sun_handle": DXFAttr(361, optional=True, dxfversion=DXF2010),
    },
)
acdb_view_group_codes = group_code_mapping(acdb_view)


@register_entity
class View(DXFEntity):
    """DXF VIEW entity"""

    DXFTYPE = "VIEW"
    DXFATTRIBS = DXFAttributes(base_class, acdb_symbol_table_record, acdb_view)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            processor.simple_dxfattribs_loader(dxf, acdb_view_group_codes)  # type: ignore
        return dxf

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        super().export_entity(tagwriter)
        if tagwriter.dxfversion > DXF12:
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_symbol_table_record.name)
            tagwriter.write_tag2(SUBCLASS_MARKER, acdb_view.name)

        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "name",
                "flags",
                "height",
                "width",
                "center",
                "direction",
                "target",
                "focal_length",
                "front_clipping",
                "back_clipping",
                "view_twist",
                "view_mode",
                "render_mode",
                "ucs",
                "ucs_origin",
                "ucs_xaxis",
                "ucs_yaxis",
                "ucs_ortho_type",
                "elevation",
                "ucs_handle",
                "base_ucs_handle",
                "camera_plottable",
                "background_handle",
                "live_selection_handle",
                "visual_style_handle",
                "sun_handle",
            ],
        )



#  Copyright (c) 2021-2022, Manfred Moitzi
#  License: MIT License

import pytest
import ezdxf
from ezdxf import explode
from ezdxf.layouts import Modelspace


@pytest.fixture(scope="module")
def msp() -> Modelspace:
    doc = ezdxf.new()
    block = doc.blocks.new("BLK1")
    block.add_line((0, 0), (1, 0))
    return doc.modelspace()


def test_virtual_entities_from_insert(msp: Modelspace):
    insert = msp.add_blockref("BLK1", (0, 0))
    result = list(explode.virtual_block_reference_entities(insert))
    assert len(result) == 1


def test_transparency_of_virtual_entities_from_insert(msp: Modelspace):
    insert = msp.add_blockref("BLK1", (0, 0))
    line = list(explode.virtual_block_reference_entities(insert))[0]
    assert line.transparency == 0.0
    assert line.dxf.hasattr("transparency") is False


def test_complex_target_coordinate_system(msp: Modelspace):
    doc = msp.doc
    block2 = doc.blocks.new("BLK2")
    insert0 = block2.add_blockref("BLK1", (0, 0))
    insert0.dxf.rotation = 30
    insert0.dxf.xscale = 7
    insert0.dxf.yscale = 3
    insert0.dxf.zscale = 1

    # this raises an internal InsertTransformationError!
    insert0.dxf.extrusion = (1, 1, 1)

    insert = msp.add_blockref("BLK2", (0, 0))
    insert.dxf.rotation = 30
    insert.dxf.xscale = 2
    insert.dxf.yscale = -5
    insert.dxf.zscale = 1

    line = list(explode.virtual_block_reference_entities(insert))[0]
    assert line.dxf.start.isclose((0, 0, 0))
    assert line.dxf.end.isclose(
        (-2.755149853494064, -18.089844737258662, 2.857738033247041)
    )


def test_explode_scaled_block_ref_containing_a_hatch(msp: Modelspace):
    doc = msp.doc
    block = doc.blocks.new("HATCH_BLK")
    hatch = block.add_hatch()
    hatch.paths.add_polyline_path([(0, 0), (1, 0), (1, 1), (0, 1)])
    hatch.set_pattern_fill("ANSI33")

    insert = msp.add_blockref("HATCH_BLK", (0, 0, 0))
    insert.scale(10, 20, 1)
    exploded_hatch = insert.explode()[0]
    assert exploded_hatch.dxf.pattern_scale == 10


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2020-2023, Manfred Moitzi
#  License: MIT License

import pytest
import platform

from ezdxf.math import BoundingBox2d
from ezdxf.fonts import fonts

TEST_FONTS = [
    ("LiberationSans-Regular.ttf", "Liberation Sans"),
    ("LiberationSerif-Regular.ttf", "Liberation Serif"),
    ("LiberationMono-Regular.ttf", "Liberation Mono"),
    ("LiberationSansNarrow-Regular.ttf", "Liberation Sans Narrow"),
    ("DejaVuSans.ttf", "DejaVu Sans"),
    ("DejaVuSerif.ttf", "DejaVu Serif"),
    ("DejaVuSansMono.ttf", "DejaVu Sans Mono"),
    ("OpenSans-Regular.ttf", "Open Sans"),
    ("OpenSansCondensed-Light.ttf", "Open Sans Condensed"),
    ("NotoSansSC-Regular.otf", "Noto Sans SC"),
]


def test_sut_font_cache_was_loaded():
    assert fonts.SUT_FONT_MANAGER_CACHE is True


@pytest.mark.parametrize("ttf,family", TEST_FONTS)
def test_if_all_test_fonts_are_available(ttf, family):
    font_face = fonts.font_manager.get_font_face(ttf)
    assert font_face.family == family, "required test-font not found"


def default_font():
    return fonts.font_manager.get_font_face("default")


def test_find_font_face_without_definition():
    assert (
        fonts.find_font_face("mozman.ttf") is default_font()
    ), "returns a default font"
    assert fonts.find_font_face(None) is default_font(), "returns a default font"


def test_find_font_face():
    r = fonts.find_font_face("LiberationSans-Regular.ttf")
    assert fonts.find_font_face("LiberationSans-Regular.ttf") == (
        "LiberationSans-Regular.ttf",
        "Liberation Sans",
        "Regular",
        400,
        5,
    )


def test_get_font_without_definition():
    default_font = fonts.font_manager.get_font_face("default")
    assert fonts.get_font_face("mozman.ttf") is default_font


def test_get_font_face_with_definition():
    assert fonts.get_font_face("Arial.ttf") is fonts.find_font_face("arial.ttf")


def test_map_shx_to_ttf():
    assert fonts.map_shx_to_ttf("TXT") == "txt_____.ttf"
    assert fonts.map_shx_to_ttf("TXT.SHX") == "txt_____.ttf"
    assert fonts.map_shx_to_ttf("txt.shx") == "txt_____.ttf"


def test_map_ttf_to_shx():
    assert fonts.map_ttf_to_shx("txt_____.ttf") == "TXT.SHX"
    assert fonts.map_ttf_to_shx("TXT_____.TTF") == "TXT.SHX"
    assert fonts.map_ttf_to_shx("xxx.ttf") is None


def test_get_font_measurement():
    assert len(fonts.get_font_measurements("LiberationSans-Regular.ttf")) == 4


def test_get_font_measurement_for_shx_fonts():
    assert len(fonts.get_font_measurements("TXT.shx")) == 4


class TestFontFace:
    def test_same_font_faces_have_equal_hash_values(self):
        f1 = fonts.FontFace("arial.ttf", "Arial")
        f2 = fonts.FontFace("arial.ttf", "Arial")
        assert hash(f1) == hash(f2)

    def test_font_face_is_italic(self):
        assert fonts.FontFace(style="italic").is_italic is True
        assert fonts.FontFace(style="oblique-italic").is_italic is True
        assert fonts.FontFace(style="").is_italic is False
        assert fonts.FontFace().is_italic is False

    def test_font_face_is_oblique(self):
        assert fonts.FontFace(style="oblique").is_oblique is True
        assert fonts.FontFace(style="oblique-italic").is_oblique is True
        assert fonts.FontFace(style="").is_oblique is False
        assert fonts.FontFace().is_oblique is False

    def test_font_face_is_bold(self):
        assert fonts.FontFace().is_bold is False
        assert fonts.FontFace(weight=300).is_bold is False
        assert fonts.FontFace(weight=500).is_bold is True
        assert fonts.FontFace(weight=700).is_bold is True
        assert fonts.FontFace(weight=900).is_bold is True

    def test_weight_str(self):
        assert fonts.FontFace(weight=-1000).weight_str == "Thin"
        assert fonts.FontFace(weight=0).weight_str == "Thin"
        assert fonts.FontFace(weight=90).weight_str == "Thin"
        assert fonts.FontFace(weight=110).weight_str == "Thin"
        assert fonts.FontFace(weight=390).weight_str == "Normal"
        assert fonts.FontFace(weight=410).weight_str == "Normal"
        assert fonts.FontFace(weight=890).weight_str == "Black"
        assert fonts.FontFace(weight=950).weight_str == "Black"
        assert fonts.FontFace(weight=3000).weight_str == "Black"

    def test_width_str(self):
        assert fonts.FontFace(width=-1000).width_str == "UltraCondensed"
        assert fonts.FontFace(width=0).width_str == "UltraCondensed"
        assert fonts.FontFace(width=1).width_str == "UltraCondensed"
        assert fonts.FontFace(width=5).width_str == "Medium"
        assert fonts.FontFace(width=9).width_str == "UltraExpanded"
        assert fonts.FontFace(width=1000).width_str == "UltraExpanded"


class TestFontMeasurements:
    @pytest.fixture
    def default(self):
        return fonts.FontMeasurements(
            baseline=1.3, cap_height=1.0, x_height=0.5, descender_height=0.25
        )

    def test_total_height(self, default):
        assert default.total_height == 1.25

    def test_scale(self, default):
        fm = default.scale(2)
        assert fm.baseline == 2.6, "expected scaled baseline"
        assert fm.total_height == 2.5, "expected scaled total height"

    def test_shift(self, default):
        fm = default.shift(1.0)
        assert fm.baseline == 2.3
        assert fm.total_height == 1.25

    def test_scale_from_baseline(self, default):
        fm = default.scale_from_baseline(desired_cap_height=2.0)
        assert fm.baseline == 1.3, "expected unchanged baseline value"
        assert fm.cap_height == 2.0
        assert fm.x_height == 1.0
        assert fm.descender_height == 0.50
        assert fm.total_height == 2.5

    def test_cap_top(self, default):
        assert default.cap_top == 2.3

    def test_x_top(self, default):
        assert default.x_top == 1.8

    def test_bottom(self, default):
        assert default.bottom == 1.05


class TestMonospaceFont:
    @pytest.fixture(scope="class")
    def mono(self):
        # special name tio create the MonospaceFont for testing
        return fonts.make_font("*monospace", 2.5, 0.75)

    def test_space_width(self, mono):
        assert mono.space_width() == pytest.approx(1.875)

    def test_text_width(self, mono):
        assert mono.text_width("1234") == pytest.approx(7.5)

    def test_text_width_ex(self, mono):
        assert mono.text_width_ex(
            "1234", cap_height=3, width_factor=1
        ) == pytest.approx(12)

    def test_text_path(self, mono):
        box = BoundingBox2d(mono.text_path("1234").control_vertices())
        assert box.size.x == pytest.approx(7.5)
        assert box.size.y == pytest.approx(2.5)

    def test_text_path_ex(self, mono):
        box = BoundingBox2d(
            mono.text_path_ex("1234", cap_height=3, width_factor=1).control_vertices()
        )
        assert box.size.x == pytest.approx(12)
        assert box.size.y == pytest.approx(3)


class TestTrueTypeFont:
    """Just test if the methods are implemented and return plausible values, don't test
    exact values, this is not a test of the rendering engine itself and results may
    change by the next revision of the font DejaVuSans.ttf or the fontTools.
    """

    @pytest.fixture(scope="class")
    def ttf(self):
        return fonts.make_font("DejaVuSans.ttf", 2.5)

    def test_space_width(self, ttf):
        assert ttf.space_width() > 1

    def test_text_width(self, ttf):
        assert ttf.text_width("1234") > 7.5

    def test_text_width_ex(self, ttf):
        assert ttf.text_width_ex("1234", cap_height=3, width_factor=2) > 19

    def test_text_path(self, ttf):
        box = BoundingBox2d(ttf.text_path("1234").control_vertices())
        assert box.size.x > 7.5
        assert box.size.y > 2.5

    def test_text_path_ex(self, ttf):
        box = BoundingBox2d(
            ttf.text_path_ex("1234", cap_height=3, width_factor=2).control_vertices()
        )
        assert box.size.x > 19
        assert box.size.y > 3


# This test works when testing only this test script in PyCharm or with pytest, but does
# not work when launching the whole test suite.
@pytest.mark.skipif(
    platform.system() != "Windows", reason="does not work on github/linux?"
)
def test_find_font_file_by_best_match():
    assert (
        fonts.find_best_match(family="Noto Sans SC").filename
        == "NotoSansSC-Regular.otf"
    )
    assert fonts.find_best_match(family="mozman") is None
    assert (
        fonts.find_best_match(family="Liberation Sans").filename
        == "LiberationSans-Regular.ttf"
    )
    assert fonts.find_best_match(family="DejaVu Sans").filename == "DejaVuSans.ttf"


# Works on Windows and Linux Mint, but not on github/Ubuntu
@pytest.mark.skipif(platform.system() != "Windows", reason="does not work on github")
def test_find_generic_font_family():
    assert fonts.find_best_match(family="serif").filename == "DejaVuSerif.ttf"
    assert fonts.find_best_match(family="sans-serif").filename == "DejaVuSans.ttf"
    assert fonts.find_best_match(family="monospace").filename == "DejaVuSansMono.ttf"


if __name__ == "__main__":
    pytest.main([__file__])



# Created: 25.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


@pytest.fixture(scope="module")
def modelspace(doc):
    return doc.modelspace()


@pytest.fixture(scope="module")
def paperspace(doc):
    return doc.layout()


def test_default_properties(modelspace):
    assert modelspace.units == 0


def test_set_units(modelspace):
    assert modelspace.units == 0
    modelspace.units = 6
    assert modelspace.units == 6
    # reset - because module scope
    modelspace.units = 0


def test_delete_entity():
    doc = ezdxf.new("R12")
    layout = doc.modelspace()
    for _ in range(5):
        layout.add_line((0, 0), (10, 0))
    lines = layout.query("LINE")
    assert 5 == len(lines)
    line3 = lines[2]
    layout.delete_entity(line3)
    assert line3.is_alive is False


def test_delete_all_entities(doc):
    paperspace = doc.layout()
    paperspace_count = len(paperspace)
    modelspace = doc.modelspace()
    modelspace_count = len(modelspace)
    for _ in range(5):
        modelspace.add_line((0, 0), (1, 1))
        paperspace.add_line((0, 0), (1, 1))

    assert modelspace_count + 5 == len(modelspace)
    assert paperspace_count + 5 == len(paperspace)

    modelspace.delete_all_entities()
    assert len(modelspace) == 0
    assert paperspace_count + 5 == len(paperspace)


def test_paper_space(paperspace):
    line = paperspace.add_line((0, 0), (1, 1))
    assert line.dxf.paperspace == 1


def test_iter_layout(doc):
    paperspace = doc.layout()
    paperspace.delete_all_entities()
    paperspace.add_line((0, 0), (1, 1))
    paperspace.add_line((0, 0), (1, 1))
    entities = list(paperspace)
    assert len(entities) == 2
    assert entities[0].is_alive is True


def test_query_entities(doc):
    paperspace = doc.layout()
    paperspace.add_line((0, 0), (1, 1), dxfattribs={"layer": "lay_lines"})
    paperspace.add_line((0, 0), (1, 1), dxfattribs={"layer": "lay_lines"})
    entities = paperspace.query('*[layer ? "lay_.*"]')
    assert len(entities) == 2
    assert entities[0].doc is doc


def test_model_space_get_layout_for_entity(doc, modelspace):
    line = modelspace.add_line((0, 0), (1, 1))
    layout = doc.layouts.get_layout_for_entity(line)
    assert modelspace is layout


def test_paper_space_get_layout_for_entity(doc):
    paperspace = doc.layout()
    line = paperspace.add_line((0, 0), (1, 1))
    layout = doc.layouts.get_layout_for_entity(line)
    assert paperspace is layout


def test_default_entity_settings(modelspace):
    line = modelspace.add_line((0, 0), (1, 1))
    assert "0" == line.dxf.layer
    assert 256 == line.dxf.color
    assert "BYLAYER" == line.dxf.linetype
    assert (0.0, 0.0, 1.0) == line.dxf.extrusion


def test_clone_dxfattribs(modelspace):
    line = modelspace.add_line((0, 0), (1, 1))
    attribs = line.dxfattribs()
    assert (
        "extrusion" not in attribs
    ), "Don't clone unset attribs with default values."


def test_invalid_layer_name(modelspace):
    with pytest.raises(ezdxf.DXFValueError):
        modelspace.add_line(
            (0, 0), (1, 1), dxfattribs={"layer": "InvalidName*"}
        )


def test_create_layout(doc):
    assert (
        len(doc.layouts) == 2
    ), "New drawing should have 1 model space and 1 paper space"

    # create a new layout
    layout = doc.layouts.new("ezdxf")
    assert "ezdxf" in doc.layouts
    assert len(layout) == 0, "New layout should contain no entities"

    with pytest.raises(ezdxf.DXFValueError):
        doc.layouts.new('invalid characters: <>/":;?*|=`')

    layout.page_setup()  # default paper setup
    assert len(layout) == 1, "missing 'main' viewport entity"


class TestAcquireLayouts:
    @pytest.fixture(scope="class")
    def doc(self):
        doc = ezdxf.new()
        doc.layouts.new("Layout2")
        doc.layouts.new("Layout3")
        doc.layouts.set_active_layout("Layout3")
        return doc

    def test_get_modelspace(self, doc):
        msp = doc.modelspace()
        assert msp.name == "Model"

    def test_layout_returns_first_layout_in_tab_order(self, doc):
        psp = doc.layout()
        assert psp.name == "Layout1"
        assert psp.is_active_paperspace is False

    def test_paperspace_returns_active_paperspace_by_default(self, doc):
        psp = doc.paperspace()
        assert psp.name == "Layout3"
        assert psp.is_active_paperspace is True

    def test_paperspace_returns_paperspace_by_name(self, doc):
        psp = doc.paperspace("Layout2")
        assert psp.name == "Layout2"
        assert psp.is_active_paperspace is False

    def test_paperspace_raises_key_error_if_not_exists(self, doc):
        with pytest.raises(KeyError):
            doc.paperspace("DoesNotExist")

    def test_paperspace_raises_key_error_for_modelspace(self, doc):
        with pytest.raises(KeyError):
            doc.paperspace("Model")



# Created: 30.04.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from io import StringIO

from ezdxf.lldxf.tags import Tags, DXFTag
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf import DXFKeyError, DXFValueError
from ezdxf.lldxf.tagwriter import TagWriter


@pytest.fixture
def xtags1():
    return ExtendedTags.from_text(XTAGS1)


def test_init_appdata(xtags1):
    assert xtags1.get_app_data("{ACAD_XDICTIONARY") is not None


def test_init_with_tags():
    tags = Tags.from_text(XTAGS1)
    xtags = ExtendedTags(tags)
    assert 3 == len(xtags.subclasses)
    assert 1 == len(xtags.xdata)


def test_init_xdata(xtags1):
    assert xtags1.get_xdata("RAK") is not None


def test_init_one_tag():
    xtags = ExtendedTags([DXFTag(0, "SECTION")])
    assert xtags.noclass[0] == (0, "SECTION")


def test_getitem(xtags1):
    assert xtags1[0] == xtags1.noclass[0]


def test_appdata_content_count(xtags1):
    xdict = xtags1.get_app_data("{ACAD_XDICTIONARY")
    assert 3 == len(xdict)


def test_appdata_content(xtags1):
    xdict = xtags1.get_app_data("{ACAD_XDICTIONARY")
    assert xdict.get_first_value(360) == "63D5"


def test_tags_skips_appdata_content(xtags1):
    with pytest.raises(DXFValueError):
        xtags1.noclass.get_first_value(360)


def test_xdata_content_count(xtags1):
    rak = xtags1.get_xdata("RAK")
    assert 17 == len(rak)


def test_tags_skips_xdata_content(xtags1):
    with pytest.raises(DXFValueError):
        xtags1.noclass.get_first_value(1000)


def test_copy(xtags1):
    stream = StringIO()
    tagwriter = TagWriter(stream)
    tagwriter.write_tags(xtags1)
    assert XTAGS1 == stream.getvalue()
    stream.close()


def test_getitem_layer(xtags1):
    assert xtags1.noclass[0] == (0, "LAYER")


def test_getitem_xdict(xtags1):
    assert xtags1.noclass[2] == (102, 0)


def test_getitem_parent(xtags1):
    assert xtags1.noclass[3] == (330, "18")


def test_get_last_item(xtags1):
    assert xtags1.noclass[-1] == (330, "18")


def test_tagscount(xtags1):
    """apdata counts as one tag and xdata counts as one tag."""
    assert 4 == len(xtags1.noclass)


def test_subclass_AcDbSymbolTableRecord(xtags1):
    subclass = xtags1.get_subclass("AcDbSymbolTableRecord")
    assert 1 == len(subclass)


def test_subclass_AcDbLayerTableRecord(xtags1):
    subclass = xtags1.get_subclass("AcDbLayerTableRecord")
    assert 8 == len(subclass)


def test_clone_is_equal(xtags1):
    clone = xtags1.clone()
    assert xtags1 is not clone
    assert xtags1.appdata is not clone.appdata
    assert xtags1.subclasses is not clone.subclasses
    assert xtags1.xdata is not clone.xdata
    assert list(xtags1) == list(clone)


def test_replace_handle(xtags1):
    xtags1.replace_handle("AA")
    assert "AA" == xtags1.get_handle()


XTAGS1 = """  0
LAYER
  5
7
102
{ACAD_XDICTIONARY
360
63D5
102
}
330
18
100
AcDbSymbolTableRecord
100
AcDbLayerTableRecord
  2
0
 70
0
 62
7
  6
CONTINUOUS
370
-3
390
8
347
805
1001
RAK
1000
{75-LÄNGENSCHNITT-14
1070
0
1070
7
1000
CONTINUOUS
1071
-3
1071
1
1005
8
1000
75-LÄNGENSCHNITT-14}
1000
{75-LÄNGENSCHNITT-2005
1070
0
1070
7
1000
CONTINUOUS
1071
-3
1071
1
1005
8
1000
75-LÄNGENSCHNITT-2005}
"""


@pytest.fixture
def xtags2():
    return ExtendedTags.from_text(XTAGS2)


def test_xdata_count(xtags2):
    assert 3 == len(xtags2.xdata)


def test_tags_count(xtags2):
    """3 xdata chunks and two 'normal' tag."""
    assert 2 == len(xtags2.noclass)


def test_xdata3_tags(xtags2):
    xdata = xtags2.get_xdata("XDATA3")
    assert xdata[0] == (1001, "XDATA3")
    assert xdata[1] == (1000, "TEXT-XDATA3")
    assert xdata[2] == (1070, 2)
    assert xdata[3] == (1070, 3)


def test_new_data(xtags2):
    xtags2.new_xdata("NEWXDATA", [(1000, "TEXT")])
    assert xtags2.has_xdata("NEWXDATA") is True

    xdata = xtags2.get_xdata("NEWXDATA")
    assert xdata[0] == (1001, "NEWXDATA")
    assert xdata[1] == (1000, "TEXT")


def test_set_new_data(xtags2):
    xtags2.new_xdata("NEWXDATA", tags=[(1000, "Extended Data String")])
    assert xtags2.has_xdata("NEWXDATA") is True

    xdata = xtags2.get_xdata("NEWXDATA")
    assert (1001, "NEWXDATA") == xdata[0]
    assert (1000, "Extended Data String") == xdata[1]


def test_append_xdata(xtags2):
    xdata = xtags2.get_xdata("MOZMAN")
    assert 4 == len(xdata)

    xdata.append(DXFTag(1000, "Extended Data String"))
    xdata = xtags2.get_xdata("MOZMAN")
    assert 5 == len(xdata)

    assert DXFTag(1000, "Extended Data String") == xdata[4]


XTAGS2 = """  0
LAYER
  5
7
1001
RAK
1000
TEXT-RAK
1070
1
1070
1
1001
MOZMAN
1000
TEXT-MOZMAN
1070
2
1070
2
1001
XDATA3
1000
TEXT-XDATA3
1070
2
1070
3
"""


@pytest.fixture
def xtags3():
    return ExtendedTags.from_text(SPECIALCASE_TEXT)


def test_read_tags(xtags3):
    subclass2 = xtags3.get_subclass("AcDbText")
    assert (100, "AcDbText") == subclass2[0]


def test_read_tags_2(xtags3):
    subclass2 = xtags3.get_subclass("AcDbText")
    assert (100, "AcDbText") == subclass2[0]
    assert (1, "Title:") == subclass2[3]


def test_read_tags_3(xtags3):
    subclass2 = xtags3.get_subclass("AcDbText", 3)
    assert (100, "AcDbText") == subclass2[0]
    assert (73, 2) == subclass2[1]


def test_key_error(xtags3):
    with pytest.raises(DXFKeyError):
        xtags3.get_subclass("AcDbText", pos=4)


def test_skip_empty_subclass(xtags3):
    xtags3.subclasses[1] = Tags()  # create empty subclass
    subclass2 = xtags3.get_subclass("AcDbText")
    assert (100, "AcDbText") == subclass2[0]


SPECIALCASE_TEXT = """  0
TEXT
5
8C9
330
6D
100
AcDbEntity
8
0
100
AcDbText
10
4.30
20
1.82
30
0.0
40
0.125
1
Title:
41
0.85
7
ARIALNARROW
100
AcDbText
73
2
"""

ACAD_REACTORS = "{ACAD_REACTORS"


@pytest.fixture
def xtags4():
    return ExtendedTags.from_text(NO_REACTORS)


def test_get_not_existing_reactor(xtags4):
    with pytest.raises(DXFValueError):
        xtags4.get_app_data(ACAD_REACTORS)


def test_new_reactors(xtags4):
    xtags4.new_app_data(ACAD_REACTORS)
    assert (102, 0) == xtags4.noclass[
        -1
    ]  # code = 102, value = index in appdata list


def test_append_not_existing_reactors(xtags4):
    xtags4.new_app_data(ACAD_REACTORS, [DXFTag(330, "DEAD")])
    reactors = xtags4.get_app_data_content(ACAD_REACTORS)
    assert 1 == len(reactors)
    assert DXFTag(330, "DEAD") == reactors[0]


def test_append_to_existing_reactors(xtags4):
    xtags4.new_app_data(ACAD_REACTORS, [DXFTag(330, "DEAD")])
    reactors = xtags4.get_app_data_content(ACAD_REACTORS)
    reactors.append(DXFTag(330, "DEAD2"))
    xtags4.set_app_data_content(ACAD_REACTORS, reactors)

    reactors = xtags4.get_app_data_content(ACAD_REACTORS)
    assert DXFTag(330, "DEAD") == reactors[0]
    assert DXFTag(330, "DEAD2") == reactors[1]


NO_REACTORS = """  0
TEXT
  5
8C9
330
6D
100
AcDbEntity
  8
0
100
AcDbText
 10
4.30
 20
1.82
 30
0.0
 40
0.125
  1
Title:
 41
0.85
  7
ARIALNARROW
"""


def test_legacy_mode():
    """Legacy mode does the same job as filter_subclass_markers()."""
    tags = ExtendedTags.from_text(LEICA_DISTO_TAGS, legacy=True)
    assert 9 == len(tags.noclass)
    assert 1 == len(tags.subclasses)
    assert tags.noclass[0] == (0, "LINE")
    assert tags.noclass[1] == (8, "LEICA_DISTO_3D")
    assert tags.noclass[-1] == (210, (0, 0, 1))


LEICA_DISTO_TAGS = """0
LINE
100
AcDbEntity
8
LEICA_DISTO_3D
62
256
6
ByLayer
5
75
100
AcDbLine
10
0.819021
20
-0.633955
30
-0.273577
11
0.753216
21
-0.582009
31
-0.276937
39
0
210
0
220
0
230
1
"""


def test_group_code_1000_outside_XDATA():
    tags = ExtendedTags(Tags.from_text(BLOCKBASEPOINTPARAMETER_CVIL_3D_2018))
    assert tags.dxftype() == "BLOCKBASEPOINTPARAMETER"
    assert len(tags.subclasses) == 6
    block_base_point_parameter = tags.get_subclass(
        "AcDbBlockBasepointParameter"
    )
    assert len(block_base_point_parameter) == 3
    assert block_base_point_parameter[0] == (100, "AcDbBlockBasepointParameter")
    assert block_base_point_parameter[1] == (1011, (0.0, 0.0, 0.0))
    assert block_base_point_parameter[2] == (1012, (0.0, 0.0, 0.0))

    block_element = tags.get_subclass("AcDbBlockElement")
    assert block_element[4] == (1071, 0)

    stream = StringIO()
    tagwriter = TagWriter(stream)
    tagwriter.write_tags(tags)
    lines = stream.getvalue()
    stream.close()
    assert len(lines.split("\n")) == len(
        BLOCKBASEPOINTPARAMETER_CVIL_3D_2018.split("\n")
    )


BLOCKBASEPOINTPARAMETER_CVIL_3D_2018 = """0
BLOCKBASEPOINTPARAMETER
5
4C25
330
4C23
100
AcDbEvalExpr
90
1
98
33
99
4
100
AcDbBlockElement
300
Base Point
98
33
99
4
1071
0
100
AcDbBlockParameter
280
1
281
0
100
AcDbBlock1PtParameter
1010
-3.108080399920343
1020
-0.9562299080084814
1030
0.0
93
0
170
0
171
0
100
AcDbBlockBasepointParameter
1011
0.0
1021
0.0
1031
0.0
1012
0.0
1022
0.0
1032
0.0
"""


def test_xrecord_with_group_code_102():
    tags = ExtendedTags(Tags.from_text(XRECORD_WITH_GROUP_CODE_102))
    assert tags.dxftype() == "XRECORD"
    assert len(tags.appdata) == 1
    assert tags.noclass[2] == (102, 0)  # 0 == index in appdata list
    assert tags.appdata[0][0] == (102, "{ACAD_REACTORS")

    xrecord = tags.get_subclass("AcDbXrecord")
    assert xrecord[2] == (102, "ACAD_ROUNDTRIP_PRE2007_TABLESTYLE")
    assert len(list(tags)) * 2 + 1 == len(
        XRECORD_WITH_GROUP_CODE_102.split("\n")
    )  # +1 == appending '\n'


XRECORD_WITH_GROUP_CODE_102 = """0
XRECORD
5
D9B071D01A0CB6A5
102
{ACAD_REACTORS
330
D9B071D01A0CB69D
102
}
330
D9B071D01A0CB69D
100
AcDbXrecord
280
 1
102
ACAD_ROUNDTRIP_PRE2007_TABLESTYLE
90
    4
91
    0
1

92
    4
93
    0
2

94
    4
95
    0
3

"""


def test_xrecord_with_long_closing_tag():
    tags = ExtendedTags(Tags.from_text(XRECORD_APP_DATA_LONG_CLOSING_TAG))
    assert tags.dxftype() == "XRECORD"
    # real app data just exists only in the base class, app data marker in AcDbXrecord are just tags, interpreted by
    # the associated application
    assert len(tags.appdata) == 1
    assert len(tags.subclasses[1]) == 35


XRECORD_APP_DATA_LONG_CLOSING_TAG = """  0
XRECORD
5
2F9
102
{ACAD_REACTORS
330
2FF
102
}
330
2FF
100
AcDbXrecord
280
1
1
AcDb_Thumbnail_Schema
102
{ATTRRECORD
341
2FA
102
USUAL_102_TAG_INSIDE_APP_DATA
2
AcDbDs::TreatedAsObjectData
280
1
291
1
102
ATTRRECORD}
102
USUAL_102_TAG_OUTSIDE_APP_DATA
102
{ATTRRECORD
341
2FB
2
AcDbDs::Legacy
280
1
291
1
102
ATTRRECORD}
2
AcDbDs::ID
280
10
91
8
102
{ATTRRECORD
341
2FC
2
AcDs:Indexable
280
1
291
1
102
ATTRRECORD}
102
{ATTRRECORD
341
2FD
2
AcDbDs::HandleAttribute
280
7
282
1
102
ATTRRECORD}
2
Thumbnail_Data
280
15
91
0
"""



# Copyright (c) 2020-2023, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterator, cast

from ezdxf import ARROWS
from ezdxf.entities import factory
from ezdxf.lldxf.const import BYBLOCK
from ezdxf.math import Vec3, fit_points_to_cad_cv

if TYPE_CHECKING:
    from ezdxf.entities import DXFGraphic, Leader, Insert, Spline, Dimension, Line


def virtual_entities(leader: Leader) -> Iterator[DXFGraphic]:
    # Source: https://atlight.github.io/formats/dxf-leader.html
    # GDAL: DXF LEADER implementation:
    # https://github.com/OSGeo/gdal/blob/master/gdal/ogr/ogrsf_frmts/dxf/ogrdxf_leader.cpp
    # LEADER DXF Reference:
    # http://help.autodesk.com/view/OARX/2018/ENU/?guid=GUID-396B2369-F89F-47D7-8223-8B7FB794F9F3
    assert leader.dxftype() == "LEADER"

    vertices = Vec3.list(leader.vertices)  # WCS
    if len(vertices) < 2:
        # This LEADER entities should be removed by the auditor if loaded or
        # ignored at exporting, if created by an ezdxf-user (log).
        raise ValueError("More than 1 vertex required.")
    dxf = leader.dxf
    doc = leader.doc

    # Some default values depend on the measurement system
    # 0/1 = imperial/metric
    if doc:
        measurement = doc.header.get("$MEASUREMENT", 0)
    else:
        measurement = 0

    # Set default styling attributes values:
    dimtad = 1
    dimgap = 0.625 if measurement else 0.0625
    dimscale = 1.0
    dimclrd = dxf.color
    dimltype = dxf.linetype
    dimlwd = dxf.lineweight
    override = None

    if doc:
        # get styling attributes from associated DIMSTYLE and/or XDATA override
        override = leader.override()
        dimtad = override.get("dimtad", dimtad)
        dimgap = override.get("dimgap", dimgap)
        dimscale = override.get("dimscale", dimscale)
        if dimscale == 0.0:  # special but unknown meaning
            dimscale = 1.0
        dimclrd = override.get("dimclrd", dimclrd)
        dimltype = override.get("dimltype", dimltype)
        dimlwd = override.get("dimlwd", dimlwd)

    text_width = dxf.text_width
    hook_line_vector = Vec3(dxf.horizontal_direction)
    has_text_annotation = dxf.annotation_type == 0

    if has_text_annotation and dxf.has_hookline:
        if dxf.hookline_direction == 1:
            hook_line_vector = -hook_line_vector
        if dimtad != 0 and text_width > 0:
            hook_line = hook_line_vector * (dimgap * dimscale + text_width)
            vertices.append(vertices[-1] + hook_line)

    dxfattribs = leader.graphic_properties()
    dxfattribs["color"] = dimclrd
    dxfattribs["linetype"] = dimltype
    dxfattribs["lineweight"] = dimlwd

    if dxfattribs.get("color") == BYBLOCK:
        dxfattribs["color"] = dxf.block_color

    if dxf.path_type == 1:  # Spline
        start_tangent = vertices[1] - vertices[0]
        end_tangent = vertices[-1] - vertices[-2]
        bspline = fit_points_to_cad_cv(vertices, tangents=[start_tangent, end_tangent])
        spline = cast("Spline", factory.new("SPLINE", doc=doc))
        spline.apply_construction_tool(bspline)
        yield spline
    else:
        attribs = dict(dxfattribs)
        prev = vertices[0]
        for vertex in vertices[1:]:
            attribs["start"] = prev
            attribs["end"] = vertex
            yield cast(
                "Line",
                factory.new(dxftype="LINE", dxfattribs=attribs, doc=doc),
            )
            prev = vertex

    if dxf.has_arrowhead and override:
        arrow_name = override.get("dimldrblk", "")
        if arrow_name is None:
            return
        size = override.get("dimasz", 2.5 if measurement else 0.1875) * dimscale
        rotation = (vertices[0] - vertices[1]).angle_deg
        if doc and arrow_name in doc.blocks:
            dxfattribs.update(
                {
                    "name": arrow_name,
                    "insert": vertices[0],
                    "rotation": rotation,
                    "xscale": size,
                    "yscale": size,
                    "zscale": size,
                }
            )
            # create a virtual block reference
            insert = cast(
                "Insert", factory.new("INSERT", dxfattribs=dxfattribs, doc=doc)
            )
            yield from insert.virtual_entities()
        else:  # render standard arrows
            yield from ARROWS.virtual_entities(
                name=arrow_name,
                insert=vertices[0],
                size=size,
                rotation=rotation,
                dxfattribs=dxfattribs,
            )



# Copyright (c) 2010-2019 Manfred Moitzi
# License: MIT License
import pytest
from io import StringIO

from ezdxf.lldxf.tagger import (
    internal_tag_compiler,
    ascii_tags_loader,
    tag_compiler,
    DXFStructureError,
)
from ezdxf.lldxf.types import strtag, DXFTag, DXFVertex
from ezdxf.math import Vec3


def test_strtag_int():
    assert "  1\n1\n" == strtag((1, 1))


def test_strtag_float():
    assert " 10\n3.1415\n" == strtag((10, 3.1415))


def test_strtag_str():
    assert "  0\nSECTION\n" == strtag((0, "SECTION"))


def test_strtag2_vector():
    assert (
        " 10\n1.0\n 20\n2.0\n 30\n3.0\n"
        == DXFVertex(10, Vec3(1, 2, 3)).dxfstr()
    )
    assert (
        " 10\n1.0\n 20\n2.0\n 30\n3.0\n"
        == DXFVertex(10, Vec3((1, 2, 3))).dxfstr()
    )
    assert (
        " 10\n1.0\n 20\n2.0\n 30\n0.0\n" == DXFVertex(10, Vec3(1, 2)).dxfstr()
    )
    assert (
        " 10\n1.0\n 20\n2.0\n 30\n0.0\n" == DXFVertex(10, Vec3((1, 2))).dxfstr()
    )


def test_int_not_skip_comments():
    tags = list(internal_tag_compiler(TAGS1))
    assert 9 == len(tags)
    assert DXFTag(999, "comment") == tags[0]


def test_int_3d_coords():
    tags = list(internal_tag_compiler(TAGS_3D_COORDS))
    assert 2 == len(tags)
    assert DXFTag(10, (100, 200, 300)) == tags[1]


def test_int_2d_coords():
    tags = list(internal_tag_compiler(TAGS_2D_COORDS))
    assert 2 == len(tags)
    assert DXFTag(10, (100, 200)) == tags[1]


def test_int_multiple_2d_coords():
    tags = list(internal_tag_compiler(TAGS_2D_COORDS2))
    assert 3 == len(tags)
    assert DXFTag(10, (100, 200)) == tags[1]
    assert DXFTag(11, (1000, 2000)) == tags[2]


def test_int_no_line_break_at_eof():
    tags = list(internal_tag_compiler(TAGS_NO_LINE_BREAK_AT_EOF))
    assert 3 == len(tags)
    assert DXFTag(10, (100, 200)) == tags[1]
    assert DXFTag(11, (1000, 2000)) == tags[2]


def test_int_float_to_int():
    with pytest.raises(ValueError):
        # Floats as int not allowed for internal tag compiler.
        list(internal_tag_compiler(FLOAT_FOR_INT_TAGS))


def test_int_no_eof():
    tags = list(internal_tag_compiler(TEST_NO_EOF))
    assert 7 == len(tags)
    assert (0, "ENDSEC") == tags[-1]


def external_tag_compiler(text):
    return tag_compiler(ascii_tags_loader(StringIO(text)))


def test_low_level_tagger_skip_comments():
    tags = list(ascii_tags_loader(StringIO("999\ncomment\n0\nEOF\n")))
    assert (0, "EOF") == tags[0]
    assert len(tags) == 1


def test_low_level_tagger_not_skip_comments():
    tags = list(
        ascii_tags_loader(
            StringIO("999\ncomment\n0\nEOF\n"), skip_comments=False
        )
    )
    assert (999, "comment") == tags[0]
    assert (0, "EOF") == tags[1]
    assert len(tags) == 2


@pytest.fixture
def reader():
    return external_tag_compiler(TEST_TAGREADER)


def test_ext_next(reader):
    assert (0, "SECTION") == next(reader)


def test_ext_to_list(reader):
    assert 8 == len(list(reader))


def test_ext_one_point_reader():
    tags = list(external_tag_compiler(POINT_TAGS))
    point_tag = tags[1]
    assert (100, 200, 300) == point_tag.value


def test_xdata_coords():
    tags = list(external_tag_compiler(XDATA_COORDS))
    assert tags[0] == (1011, (100, 200, 300))
    assert len(tags) == 1


def test_ext_read_2D_points():
    stri = internal_tag_compiler(POINT_2D_TAGS)
    tags = list(stri)
    tag = tags[0]  # 2D point
    assert (100, 200) == tag.value
    tag = tags[1]  # check mark
    assert "check mark 1" == tag.value
    tag = tags[2]  # 3D point
    assert (100, 200, 300) == tag.value
    tag = tags[3]  # check mark
    assert "check mark 2" == tag.value


def test_ext_error_tag():
    tags = list(external_tag_compiler(TAGS_WITH_ERROR))
    assert 1 == len(tags)


def test_ext_float_to_int():
    # Floats as int allowed for external tag compiler (thx ProE).
    assert list(external_tag_compiler(FLOAT_FOR_INT_TAGS))[0] == (71, 1)


def test_ext_coord_error_tag():
    with pytest.raises(DXFStructureError):
        list(external_tag_compiler(TAGS_WITH_COORD_ERROR))


def test_polyline_with_xdata():
    tags = list(tag_compiler(ascii_tags_loader(StringIO(POLYLINE_WITH_XDATA))))
    assert len(tags) == 49


TAGS1 = """999
comment
  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  0
ENDSEC
  0
EOF
"""

TAGS_3D_COORDS = """  9
$EXTMIN
 10
100
 20
200
 30
300
"""

XDATA_COORDS = """1011
100
1021
200
1031
300
"""

TAGS_2D_COORDS = """  9
$EXTMIN
 10
100
 20
200
"""

TAGS_2D_COORDS2 = """  9
$EXTMIN
 10
100
 20
200
 11
1000
 21
2000
"""

TAGS_NO_LINE_BREAK_AT_EOF = """  9
$EXTMIN
 10
100
 20
200
 11
1000
 21
2000"""

TEST_TAGREADER = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  0
ENDSEC
  0
EOF
"""

TEST_NO_EOF = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  0
ENDSEC
"""

TEST_TAGREADER_COMMENTS = """999
Comment0
  0
SECTION
  2
HEADER
  9
$ACADVER
999
Comment1
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  0
ENDSEC
  0
EOF
"""

POINT_TAGS = """  9
$EXTMIN
 10
100
 20
200
 30
300
"""

POINT_2D_TAGS = """ 10
100
 20
200
  9
check mark 1
 10
100
 20
200
 30
300
  9
check mark 2
"""

FLOAT_FOR_INT_TAGS = """  71
1.0
"""

TAGS_WITH_ERROR = """  9
$EXTMIN
 10
100
 20
"""

TAGS_WITH_COORD_ERROR = """  9
$EXTMIN
 20
100
 10
100
 40
1.0  
"""

POLYLINE_WITH_XDATA = """  0
POLYLINE
  5
2A
  8
T-POLYFACE-3DS
 62
     3
 66
     1
 10
0.0
 20
0.0
 30
0.0
 70
    64
 71
     8
 72
    12
1001
AVE_FINISH
1002
{
1070
     0
1005
0
1002
}
1001
AVE_ENTITY_MATERIAL
1002
{
1000

1002
{
1071
        0
1070
     0
1070
     0
1002
{
1070
     0
1070
     0
1070
     0
1040
0.0
1002
}
1070
     0
1070
     0
1002
{
1002
}
1002
}
1002
{
1002
}
1002
{
1002
}
1011
0.0
1021
0.0
1031
0.0
1021
0.0
1031
0.0
1011
1.0
1021
0.0
1031
0.0
1021
0.0
1031
0.0
1011
0.0
1021
0.0
1031
0.0
1021
1.0
1031
0.0
1011
0.0
1021
0.0
1031
0.0
1021
0.0
1031
1.0
1002
}
"""



# Created: 12.03.2011, 2018 rewritten for pytest
# Copyright (C) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture(scope="module")
def doc():
    return ezdxf.new()


def test_duplicate_simple_entity(doc):
    msp = doc.modelspace()
    circle = msp.add_circle(
        center=(2, 3), radius=1.5, dxfattribs={"layer": "test", "color": 4}
    )
    new_circle = doc.entitydb.duplicate_entity(circle)
    assert circle.dxf.handle != new_circle.dxf.handle, "expected new handle"
    assert new_circle.dxf.center == (2, 3)
    assert new_circle.dxf.radius == 1.5
    assert new_circle.dxf.layer == "test"
    assert new_circle.dxf.color == 4
    assert new_circle.dxf.owner is None  # undefined owner/layout
    with pytest.raises(ezdxf.DXFKeyError):
        doc.layouts.get_layout_for_entity(new_circle)


def test_duplicate_polyline_entity(doc):
    msp = doc.modelspace()
    polyline = msp.add_polyline3d(
        points=[(1, 1, 1), (3, 2, -1), (7, 4, 4)],
        dxfattribs={"layer": "test", "color": 4},
    )
    start_len = len(doc.entitydb)
    new_polyline = doc.entitydb.duplicate_entity(polyline)

    assert len(doc.entitydb) == start_len + 5  # POLYLINE, 3x VERTEX, 1x SEQEND
    assert polyline.dxf.handle != new_polyline.dxf.handle, "expected new handle"
    assert new_polyline.dxf.layer == "test"
    assert new_polyline.dxf.color == 4
    assert new_polyline.dxf.owner is None  # undefined owner/layout

    assert len(polyline.vertices) == len(new_polyline.vertices)
    for v1, v2 in zip(polyline.vertices, new_polyline.vertices):
        assert v1.dxf.handle != v2.dxf.handle, "expected new handle"
        assert v1.dxf.location == v2.dxf.location


def test_duplicate_insert_with_attribs_entity(doc):
    msp = doc.modelspace()
    insert = msp.add_blockref(name="", insert=(3, 4))
    insert.add_attrib("TAG1", "content1", insert=(5, 6))
    insert.add_attrib("TAG2", "content2", insert=(6, 6))
    start_len = len(doc.entitydb)
    new_insert = doc.entitydb.duplicate_entity(insert)

    assert len(doc.entitydb) == start_len + 4  # INSERT, 2x ATTRIB, 1x SEQEND
    assert insert.dxf.handle != new_insert.dxf.handle, "expected new handle"

    assert new_insert.dxf.owner is None  # undefined owner/layout

    assert len(insert.attribs) == len(new_insert.attribs)
    for a1, a2 in zip(insert.attribs, new_insert.attribs):
        assert a1.dxf.handle != a2.dxf.handle, "expected new handle"
        assert a1.dxf.tag == a2.dxf.tag
        assert a1.dxf.text == a2.dxf.text



#  Copyright (c) 2021, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.addons.drawing.gfxproxy import DXFGraphicProxy
from ezdxf.entities import DXFObject, factory
from ezdxf.lldxf.extendedtags import ExtendedTags
from ezdxf.protocols import SupportsVirtualEntities, query_virtual_entities
from ezdxf.entities.copy import CopyNotSupported

class FakeObject(DXFObject):
    # does not support any graphical DXF attributes like layer, linetype ...
    DXFTYPE = "FAKE_OBJECT"


ENTITIES = [0, 2, 3]


class VirtualEntitiesMethod(DXFObject):
    def virtual_entities(self):
        return ENTITIES


class VirtualEntitiesProtocol(DXFObject):
    def __virtual_entities__(self):
        return ENTITIES


class TestDXFGraphicProxy:
    @pytest.fixture
    def proxy(self):
        return DXFGraphicProxy(FakeObject())

    def test_dxf_type(self, proxy):
        assert proxy.dxftype() == "FAKE_OBJECT"

    def test_get_dxfattribs(self, proxy):
        assert proxy.dxf.layer == "0", "expected default layer"

    def test_setting_dxfattribs_does_no_alter_wrapped_entity(self, proxy):
        proxy.dxf.layer = "TEST"
        assert proxy.dxf.layer == "TEST", "expected layer changed"
        assert (
            proxy.entity.dxf.hasattr("layer") is False
        ), "wrapped object does not support 'layer'"

    def test_copy_raises_cpy_not_supported_exception(self, proxy):
        with pytest.raises(CopyNotSupported):
            proxy.copy()

    def test_supports_virtual_entities_protocol(self, proxy):
        assert isinstance(proxy, SupportsVirtualEntities)
        assert len(query_virtual_entities(proxy)) == 0

    def test_supports_virtual_entities_method_of_wrapped_entity(self):
        proxy = DXFGraphicProxy(VirtualEntitiesMethod())
        assert list(query_virtual_entities(proxy)) == ENTITIES

    def test_supports_virtual_entities_protocol_of_wrapped_entity(self):
        proxy = DXFGraphicProxy(VirtualEntitiesProtocol())
        assert list(query_virtual_entities(proxy)) == ENTITIES


def test_support_for_proxy_graphic_stored_in_acdb_entity():
    tag_storage = factory.load(ExtendedTags.from_text(DXF_STORAGE))
    assert tag_storage.dxftype() == "TAG_STORAGE"
    proxy = DXFGraphicProxy(tag_storage)
    # reusing data from test suite 252
    types = [e.dxftype() for e in query_virtual_entities(proxy)]
    assert types == [
        "POLYLINE",
        "POLYLINE",
        "ARC",
        "ARC",
        "POLYLINE",
        "POLYLINE",
    ]


DXF_STORAGE = """0
TAG_STORAGE
5
FEFE
100
AcDbEntity
8
0
160
728
310
D80200000D0000000C00000010000000050000000C0000000E000000000100000C00000012000000FF7F0000100000002200000000000000000000000C00000010000000050000000C00000012000000000000000C0000000E000000010000006C00000007000000040000008705D40EB308C2404F852AD01EBFC340000000
310
0000000000188DCA03B308C240BF872AD01EA6C34000000000000000003988CA03B33AC240E19617BA1EA6C3400000000000000000A800D40EB33AC240719417BA1EBFC34000000000000000006C000000070000000400000043041686B408C2407A322AD01E11C7400000000000000000B27C1F91B408C2400A302AD01E2A
310
C7400000000000000000D3771F91B43AC2402C3F17BA1E2AC740000000000000000064FF1586B43AC2409C4117BA1E11C74000000000000000006400000004000000A800D40EB33AC240719417BA1EBFC34000000000000000000000000000908A4000000000000000000000000000000000010000000000F0BFD0A43C4F34
310
41DC3EADE1FCFFFFFFEF3F0000000000000000182D4454FB21F93F00000000640000000400000064FF1586B43AC2409C4117BA1E11C74000000000000000000000000000908A4000000000000000000000000000000000010000000000F03FD0A43C4F3441DCBEADE1FCFFFFFFEFBF0000000000000000182D4454FB21F93F
310
000000006C0000000700000004000000A800D40EB33AC240719417BA1EBFC3400000000000000000A36A8A18B33AC2404D9217BA1ED5C340000000000000000038418A18B3E3C340EE9276FE1DD5C34000000000000000003DD7D30EB3E3C340139576FE1DBFC34000000000000000006C000000070000000400000064FF15
310
86B43AC2409C4117BA1E11C740000000000000000069955F7CB43AC240C14317BA1EFBC6400000000000000000FE6B5F7CB4E3C340634476FE1DFBC6400000000000000000F9D51586B4E3C3403E4276FE1D11C7400000000000000000
100
AcDbTagStorage
"""

if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from ezdxf.math import (
    ApproxParamT,
    Bezier3P,
    Bezier4P,
    BSpline,
    open_uniform_knot_vector,
    Vec2,
)


class TestGenericFeatures:
    @pytest.fixture
    def curve(self):
        return Bezier3P(Vec2.list([(0, 0), (1, 2), (2, 4)]))

    def test_access_to_construction_polyline(self, curve):
        approx = ApproxParamT(curve, segments=100)
        assert approx.polyline.length == pytest.approx(4.472135954999586)

    def test_get_max_t(self, curve):
        approx = ApproxParamT(curve, segments=100)
        assert approx.max_t == 1.0


class TestQuadraticBezier:
    @pytest.fixture(scope="class")
    def approx(self):
        return ApproxParamT(Bezier3P(Vec2.list([(0, 0), (1, 2), (2, 4)])))

    def test_start_param(self, approx):
        assert approx.param_t(0) == pytest.approx(0.0)

    def test_end_param(self, approx):
        assert approx.param_t(approx.polyline.length) == pytest.approx(1.0)

    def test_certain_distance(self, approx):
        t = 0.4472135954999581
        assert approx.param_t(2.0) == pytest.approx(t)
        assert approx.distance(t) == pytest.approx(2.0)


class TestCubicBezier:
    @pytest.fixture(scope="class")
    def approx(self):
        return ApproxParamT(Bezier4P(Vec2.list([(0, 0), (1, 2), (2, 4), (3, 1)])))

    def test_start_param(self, approx):
        assert approx.param_t(0) == pytest.approx(0.0)

    def test_end_param(self, approx):
        assert approx.param_t(approx.polyline.length) == pytest.approx(1.0)

    def test_certain_distance(self, approx):
        t = 0.31949832059570293
        assert approx.param_t(2.0) == pytest.approx(t)
        assert approx.distance(t) == pytest.approx(2.0)


class TestCubicSpline:
    @pytest.fixture(scope="class")
    def spline(self):
        points = [(0, 0), (1, 2), (2, 4), (3, 1), (4, 2)]
        # by default knot values are normalized in the range [0, 1],
        # this creates a not normalized knot vector in the range [0, 2]:
        knots = open_uniform_knot_vector(count=len(points), order=4)
        return BSpline(
            [(0, 0), (1, 2), (2, 4), (3, 1), (4, 2)],
            knots=knots,
        )

    @pytest.fixture(scope="class")
    def approx(self, spline):
        return ApproxParamT(spline, max_t=spline.max_t)

    def test_max_t_is_not_1(self, spline):
        """This class should test a different parameter range!"""
        assert spline.max_t == pytest.approx(2.0)

    def test_start_param(self, approx):
        assert approx.param_t(0) == pytest.approx(0.0)
        assert approx.distance(0.0) == pytest.approx(0.0)

    def test_end_param(self, approx):
        length = approx.polyline.length
        assert approx.param_t(length) == pytest.approx(2.0)
        assert approx.distance(2.0) == pytest.approx(length)

    def test_certain_distance(self, approx):
        t = 0.6692720686599521
        assert approx.param_t(3.0) == pytest.approx(t)
        assert approx.distance(t) == pytest.approx(3.0)


if __name__ == "__main__":
    pytest.main([__file__])
